# Copyright 2014 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.common import exceptions as p_exc
from vmware_nsx.common import nsx_utils
from vmware_nsx.db import lsn_db
from vmware_nsx.dhcp_meta import constants as const
from vmware_nsx.nsxlib.mh import lsn as lsn_api
from vmware_nsx.nsxlib.mh import switch as switch_api

LOG = logging.getLogger(__name__)

META_CONF = 'metadata-proxy'
DHCP_CONF = 'dhcp'


lsn_opts = [
    cfg.BoolOpt('sync_on_missing_data', default=False,
                help=_('Pull LSN information from NSX in case it is missing '
                       'from the local data store. This is useful to rebuild '
                       'the local store in case of server recovery.'))
]


def register_lsn_opts(config):
    config.CONF.register_opts(lsn_opts, "NSX_LSN")


class LsnManager(object):
    """Manage LSN entities associated with networks."""

    def __init__(self, plugin):
        self.plugin = plugin

    @property
    def cluster(self):
        return self.plugin.cluster

    def lsn_exists(self, context, network_id):
        """Return True if a Logical Service Node exists for the network."""
        return self.lsn_get(
            context, network_id, raise_on_err=False) is not None

    def lsn_get(self, context, network_id, raise_on_err=True):
        """Retrieve the LSN id associated to the network."""
        try:
            return lsn_api.lsn_for_network_get(self.cluster, network_id)
        except (n_exc.NotFound, api_exc.NsxApiException):
            if raise_on_err:
                LOG.error('Unable to find Logical Service Node for '
                          'network %s.',
                          network_id)
                raise p_exc.LsnNotFound(entity='network',
                                        entity_id=network_id)
            else:
                LOG.warning('Unable to find Logical Service Node for '
                            'the requested network %s.',
                            network_id)

    def lsn_create(self, context, network_id):
        """Create a LSN associated to the network."""
        try:
            return lsn_api.lsn_for_network_create(self.cluster, network_id)
        except api_exc.NsxApiException:
            err_msg = _('Unable to create LSN for network %s') % network_id
            raise p_exc.NsxPluginException(err_msg=err_msg)

    def lsn_delete(self, context, lsn_id):
        """Delete a LSN given its id."""
        try:
            lsn_api.lsn_delete(self.cluster, lsn_id)
        except (n_exc.NotFound, api_exc.NsxApiException):
            LOG.warning('Unable to delete Logical Service Node %s',
                        lsn_id)

    def lsn_delete_by_network(self, context, network_id):
        """Delete a LSN associated to the network."""
        lsn_id = self.lsn_get(context, network_id, raise_on_err=False)
        if lsn_id:
            self.lsn_delete(context, lsn_id)

    def lsn_port_get(self, context, network_id, subnet_id, raise_on_err=True):
        """Retrieve LSN and LSN port for the network and the subnet."""
        lsn_id = self.lsn_get(context, network_id, raise_on_err=raise_on_err)
        if lsn_id:
            try:
                lsn_port_id = lsn_api.lsn_port_by_subnet_get(
                    self.cluster, lsn_id, subnet_id)
            except (n_exc.NotFound, api_exc.NsxApiException):
                if raise_on_err:
                    LOG.error('Unable to find Logical Service Node Port '
                              'for LSN %(lsn_id)s and subnet '
                              '%(subnet_id)s',
                              {'lsn_id': lsn_id, 'subnet_id': subnet_id})
                    raise p_exc.LsnPortNotFound(lsn_id=lsn_id,
                                                entity='subnet',
                                                entity_id=subnet_id)
                else:
                    LOG.warning('Unable to find Logical Service Node Port '
                                'for LSN %(lsn_id)s and subnet '
                                '%(subnet_id)s',
                                {'lsn_id': lsn_id, 'subnet_id': subnet_id})
                return (lsn_id, None)
            else:
                return (lsn_id, lsn_port_id)
        else:
            return (None, None)

    def lsn_port_get_by_mac(self, context, network_id, mac, raise_on_err=True):
        """Retrieve LSN and LSN port given network and mac address."""
        lsn_id = self.lsn_get(context, network_id, raise_on_err=raise_on_err)
        if lsn_id:
            try:
                lsn_port_id = lsn_api.lsn_port_by_mac_get(
                    self.cluster, lsn_id, mac)
            except (n_exc.NotFound, api_exc.NsxApiException):
                if raise_on_err:
                    LOG.error('Unable to find Logical Service Node Port '
                              'for LSN %(lsn_id)s and mac address '
                              '%(mac)s',
                              {'lsn_id': lsn_id, 'mac': mac})
                    raise p_exc.LsnPortNotFound(lsn_id=lsn_id,
                                                entity='MAC',
                                                entity_id=mac)
                else:
                    LOG.warning('Unable to find Logical Service Node '
                                'Port for LSN %(lsn_id)s and mac address '
                                '%(mac)s',
                                {'lsn_id': lsn_id, 'mac': mac})
                return (lsn_id, None)
            else:
                return (lsn_id, lsn_port_id)
        else:
            return (None, None)

    def lsn_port_create(self, context, lsn_id, subnet_info):
        """Create and return LSN port for associated subnet."""
        try:
            return lsn_api.lsn_port_create(self.cluster, lsn_id, subnet_info)
        except n_exc.NotFound:
            raise p_exc.LsnNotFound(entity='', entity_id=lsn_id)
        except api_exc.NsxApiException:
            err_msg = _('Unable to create port for LSN  %s') % lsn_id
            raise p_exc.NsxPluginException(err_msg=err_msg)

    def lsn_port_delete(self, context, lsn_id, lsn_port_id):
        """Delete a LSN port from the Logical Service Node."""
        try:
            lsn_api.lsn_port_delete(self.cluster, lsn_id, lsn_port_id)
        except (n_exc.NotFound, api_exc.NsxApiException):
            LOG.warning('Unable to delete LSN Port %s', lsn_port_id)

    def lsn_port_dispose(self, context, network_id, mac_address):
        """Delete a LSN port given the network and the mac address."""
        lsn_id, lsn_port_id = self.lsn_port_get_by_mac(
            context, network_id, mac_address, raise_on_err=False)
        if lsn_port_id:
            self.lsn_port_delete(context, lsn_id, lsn_port_id)
            if mac_address == const.METADATA_MAC:
                try:
                    lswitch_port_id = switch_api.get_port_by_neutron_tag(
                        self.cluster, network_id,
                        const.METADATA_PORT_ID)['uuid']
                    switch_api.delete_port(
                        self.cluster, network_id, lswitch_port_id)
                except (n_exc.PortNotFoundOnNetwork,
                        api_exc.NsxApiException):
                    LOG.warning("Metadata port not found while attempting "
                                "to delete it from network %s",
                                network_id)
        else:
            LOG.warning("Unable to find Logical Services Node "
                        "Port with MAC %s", mac_address)

    def lsn_port_dhcp_setup(
        self, context, network_id, port_id, port_data, subnet_config=None):
        """Connect network to LSN via specified port and port_data."""
        try:
            lsn_id = None
            switch_id = nsx_utils.get_nsx_switch_ids(
                context.session, self.cluster, network_id)[0]
            lswitch_port_id = switch_api.get_port_by_neutron_tag(
                self.cluster, switch_id, port_id)['uuid']
            lsn_id = self.lsn_get(context, network_id)
            lsn_port_id = self.lsn_port_create(context, lsn_id, port_data)
        except (n_exc.NotFound, p_exc.NsxPluginException):
            raise p_exc.PortConfigurationError(
                net_id=network_id, lsn_id=lsn_id, port_id=port_id)
        else:
            try:
                lsn_api.lsn_port_plug_network(
                    self.cluster, lsn_id, lsn_port_id, lswitch_port_id)
            except p_exc.LsnConfigurationConflict:
                self.lsn_port_delete(context, lsn_id, lsn_port_id)
                raise p_exc.PortConfigurationError(
                    net_id=network_id, lsn_id=lsn_id, port_id=port_id)
            if subnet_config:
                self.lsn_port_dhcp_configure(
                    context, lsn_id, lsn_port_id, subnet_config)
            else:
                return (lsn_id, lsn_port_id)

    def lsn_port_metadata_setup(self, context, lsn_id, subnet):
        """Connect subnet to specified LSN."""
        data = {
            "mac_address": const.METADATA_MAC,
            "ip_address": subnet['cidr'],
            "subnet_id": subnet['id']
        }
        network_id = subnet['network_id']
        tenant_id = subnet['tenant_id']
        lswitch_port_id = None
        try:
            switch_id = nsx_utils.get_nsx_switch_ids(
                context.session, self.cluster, network_id)[0]
            lswitch_port_id = switch_api.create_lport(
                self.cluster, switch_id, tenant_id,
                const.METADATA_PORT_ID, const.METADATA_PORT_NAME,
                const.METADATA_DEVICE_ID, True)['uuid']
            lsn_port_id = self.lsn_port_create(context, lsn_id, data)
        except (n_exc.NotFound, p_exc.NsxPluginException,
                api_exc.NsxApiException):
            raise p_exc.PortConfigurationError(
                net_id=network_id, lsn_id=lsn_id, port_id=lswitch_port_id)
        else:
            try:
                lsn_api.lsn_port_plug_network(
                    self.cluster, lsn_id, lsn_port_id, lswitch_port_id)
            except p_exc.LsnConfigurationConflict:
                self.lsn_port_delete(self.cluster, lsn_id, lsn_port_id)
                switch_api.delete_port(
                    self.cluster, network_id, lswitch_port_id)
                raise p_exc.PortConfigurationError(
                    net_id=network_id, lsn_id=lsn_id, port_id=lsn_port_id)

    def lsn_port_dhcp_configure(self, context, lsn_id, lsn_port_id, subnet):
        """Enable/disable dhcp services with the given config options."""
        is_enabled = subnet["enable_dhcp"]
        dhcp_options = {
            "domain_name": cfg.CONF.NSX_DHCP.domain_name,
            "default_lease_time": cfg.CONF.NSX_DHCP.default_lease_time,
        }
        dns_servers = cfg.CONF.NSX_DHCP.extra_domain_name_servers or []
        dns_servers.extend(subnet["dns_nameservers"])
        if subnet['gateway_ip']:
            dhcp_options["routers"] = subnet["gateway_ip"]
        if dns_servers:
            dhcp_options["domain_name_servers"] = ",".join(dns_servers)
        if subnet["host_routes"]:
            dhcp_options["classless_static_routes"] = (
                ",".join(subnet["host_routes"])
            )
        try:
            lsn_api.lsn_port_dhcp_configure(
                self.cluster, lsn_id, lsn_port_id, is_enabled, dhcp_options)
        except (n_exc.NotFound, api_exc.NsxApiException):
            err_msg = (_('Unable to configure dhcp for Logical Service '
                         'Node %(lsn_id)s and port %(lsn_port_id)s')
                       % {'lsn_id': lsn_id, 'lsn_port_id': lsn_port_id})
            LOG.error(err_msg)
            raise p_exc.NsxPluginException(err_msg=err_msg)

    def lsn_metadata_configure(self, context, subnet_id, is_enabled):
        """Configure metadata service for the specified subnet."""
        subnet = self.plugin.get_subnet(context, subnet_id)
        network_id = subnet['network_id']
        meta_conf = cfg.CONF.NSX_METADATA
        metadata_options = {
            'metadata_server_ip': meta_conf.metadata_server_address,
            'metadata_server_port': meta_conf.metadata_server_port,
            'metadata_proxy_shared_secret': meta_conf.metadata_shared_secret
        }
        try:
            lsn_id = self.lsn_get(context, network_id)
            lsn_api.lsn_metadata_configure(
                self.cluster, lsn_id, is_enabled, metadata_options)
        except (p_exc.LsnNotFound, api_exc.NsxApiException):
            err_msg = (_('Unable to configure metadata '
                         'for subnet %s') % subnet_id)
            LOG.error(err_msg)
            raise p_exc.NsxPluginException(err_msg=err_msg)
        if is_enabled:
            try:
                # test that the lsn port exists
                self.lsn_port_get(context, network_id, subnet_id)
            except p_exc.LsnPortNotFound:
                # this might happen if subnet had dhcp off when created
                # so create one, and wire it
                self.lsn_port_metadata_setup(context, lsn_id, subnet)
        else:
            self.lsn_port_dispose(context, network_id, const.METADATA_MAC)

    def _lsn_port_host_conf(self, context, network_id, subnet_id, data, hdlr):
        lsn_id, lsn_port_id = self.lsn_port_get(
            context, network_id, subnet_id, raise_on_err=False)
        try:
            if lsn_id and lsn_port_id:
                hdlr(self.cluster, lsn_id, lsn_port_id, data)
        except (n_exc.NotFound, api_exc.NsxApiException):
            LOG.error('Error while configuring LSN '
                      'port %s', lsn_port_id)
            raise p_exc.PortConfigurationError(
                net_id=network_id, lsn_id=lsn_id, port_id=lsn_port_id)

    def lsn_port_dhcp_host_add(self, context, network_id, subnet_id, host):
        """Add dhcp host entry to LSN port configuration."""
        self._lsn_port_host_conf(context, network_id, subnet_id, host,
                                 lsn_api.lsn_port_dhcp_host_add)

    def lsn_port_dhcp_host_remove(self, context, network_id, subnet_id, host):
        """Remove dhcp host entry from LSN port configuration."""
        self._lsn_port_host_conf(context, network_id, subnet_id, host,
                                 lsn_api.lsn_port_dhcp_host_remove)

    def lsn_port_meta_host_add(self, context, network_id, subnet_id, host):
        """Add dhcp host entry to LSN port configuration."""
        self._lsn_port_host_conf(context, network_id, subnet_id, host,
                                 lsn_api.lsn_port_metadata_host_add)

    def lsn_port_meta_host_remove(self, context, network_id, subnet_id, host):
        """Remove dhcp host entry from LSN port configuration."""
        self._lsn_port_host_conf(context, network_id, subnet_id, host,
                                 lsn_api.lsn_port_metadata_host_remove)

    def lsn_port_update(
        self, context, network_id, subnet_id, dhcp=None, meta=None):
        """Update the specified configuration for the LSN port."""
        if not dhcp and not meta:
            return
        try:
            lsn_id, lsn_port_id = self.lsn_port_get(
                context, network_id, subnet_id, raise_on_err=False)
            if dhcp and lsn_id and lsn_port_id:
                lsn_api.lsn_port_host_entries_update(
                    self.cluster, lsn_id, lsn_port_id, DHCP_CONF, dhcp)
            if meta and lsn_id and lsn_port_id:
                lsn_api.lsn_port_host_entries_update(
                    self.cluster, lsn_id, lsn_port_id, META_CONF, meta)
        except api_exc.NsxApiException:
            raise p_exc.PortConfigurationError(
                net_id=network_id, lsn_id=lsn_id, port_id=lsn_port_id)


class PersistentLsnManager(LsnManager):
    """Add local persistent state to LSN Manager."""

    def __init__(self, plugin):
        super(PersistentLsnManager, self).__init__(plugin)
        self.sync_on_missing = cfg.CONF.NSX_LSN.sync_on_missing_data

    def lsn_get(self, context, network_id, raise_on_err=True):
        try:
            obj = lsn_db.lsn_get_for_network(
                context, network_id, raise_on_err=raise_on_err)
            return obj.lsn_id if obj else None
        except p_exc.LsnNotFound:
            with excutils.save_and_reraise_exception() as ctxt:
                ctxt.reraise = False
                if self.sync_on_missing:
                    lsn_id = super(PersistentLsnManager, self).lsn_get(
                        context, network_id, raise_on_err=raise_on_err)
                    self.lsn_save(context, network_id, lsn_id)
                    return lsn_id
                if raise_on_err:
                    ctxt.reraise = True

    def lsn_save(self, context, network_id, lsn_id):
        """Save LSN-Network mapping to the DB."""
        try:
            lsn_db.lsn_add(context, network_id, lsn_id)
        except db_exc.DBError:
            err_msg = _('Unable to save LSN for network %s') % network_id
            LOG.exception(err_msg)
            raise p_exc.NsxPluginException(err_msg=err_msg)

    def lsn_create(self, context, network_id):
        lsn_id = super(PersistentLsnManager,
                       self).lsn_create(context, network_id)
        try:
            self.lsn_save(context, network_id, lsn_id)
        except p_exc.NsxPluginException:
            with excutils.save_and_reraise_exception():
                super(PersistentLsnManager, self).lsn_delete(context, lsn_id)
        return lsn_id

    def lsn_delete(self, context, lsn_id):
        lsn_db.lsn_remove(context, lsn_id)
        super(PersistentLsnManager, self).lsn_delete(context, lsn_id)

    def lsn_port_get(self, context, network_id, subnet_id, raise_on_err=True):
        try:
            obj = lsn_db.lsn_port_get_for_subnet(
                context, subnet_id, raise_on_err=raise_on_err)
            return (obj.lsn_id, obj.lsn_port_id) if obj else (None, None)
        except p_exc.LsnPortNotFound:
            with excutils.save_and_reraise_exception() as ctxt:
                ctxt.reraise = False
                if self.sync_on_missing:
                    lsn_id, lsn_port_id = (
                        super(PersistentLsnManager, self).lsn_port_get(
                            context, network_id, subnet_id,
                            raise_on_err=raise_on_err))
                    mac_addr = lsn_api.lsn_port_info_get(
                        self.cluster, lsn_id, lsn_port_id)['mac_address']
                    self.lsn_port_save(
                        context, lsn_port_id, subnet_id, mac_addr, lsn_id)
                    return (lsn_id, lsn_port_id)
                if raise_on_err:
                    ctxt.reraise = True

    def lsn_port_get_by_mac(self, context, network_id, mac, raise_on_err=True):
        try:
            obj = lsn_db.lsn_port_get_for_mac(
                context, mac, raise_on_err=raise_on_err)
            return (obj.lsn_id, obj.lsn_port_id) if obj else (None, None)
        except p_exc.LsnPortNotFound:
            with excutils.save_and_reraise_exception() as ctxt:
                ctxt.reraise = False
                if self.sync_on_missing:
                    lsn_id, lsn_port_id = (
                        super(PersistentLsnManager, self).lsn_port_get_by_mac(
                            context, network_id, mac,
                            raise_on_err=raise_on_err))
                    subnet_id = lsn_api.lsn_port_info_get(
                        self.cluster, lsn_id, lsn_port_id).get('subnet_id')
                    self.lsn_port_save(
                        context, lsn_port_id, subnet_id, mac, lsn_id)
                    return (lsn_id, lsn_port_id)
                if raise_on_err:
                    ctxt.reraise = True

    def lsn_port_save(self, context, lsn_port_id, subnet_id, mac_addr, lsn_id):
        """Save LSN Port information to the DB."""
        try:
            lsn_db.lsn_port_add_for_lsn(
                context, lsn_port_id, subnet_id, mac_addr, lsn_id)
        except db_exc.DBError:
            err_msg = _('Unable to save LSN port for subnet %s') % subnet_id
            LOG.exception(err_msg)
            raise p_exc.NsxPluginException(err_msg=err_msg)

    def lsn_port_create(self, context, lsn_id, subnet_info):
        lsn_port_id = super(PersistentLsnManager,
                            self).lsn_port_create(context, lsn_id, subnet_info)
        try:
            self.lsn_port_save(context, lsn_port_id, subnet_info['subnet_id'],
                               subnet_info['mac_address'], lsn_id)
        except p_exc.NsxPluginException:
            with excutils.save_and_reraise_exception():
                super(PersistentLsnManager, self).lsn_port_delete(
                    context, lsn_id, lsn_port_id)
        return lsn_port_id

    def lsn_port_delete(self, context, lsn_id, lsn_port_id):
        lsn_db.lsn_port_remove(context, lsn_port_id)
        super(PersistentLsnManager, self).lsn_port_delete(
            context, lsn_id, lsn_port_id)



# Copyright (c) 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class L3NotifyAPI(object):
    """Dummy driver for L3 notifcations - no need - no L3 agenets."""

    # We need this driver as this code is invoked from the L3 mixin code.

    def agent_updated(self, context, admin_state_up, host):
        pass

    def router_deleted(self, context, router_id):
        pass

    def routers_updated(self, context, router_ids, operation=None, data=None,
                        shuffle_agents=False, schedule_routers=True):
        pass

    def add_arp_entry(self, context, router_id, arp_table, operation=None):
        pass

    def del_arp_entry(self, context, router_id, arp_table, operation=None):
        pass

    def delete_fipnamespace_for_ext_net(self, context, ext_net_id):
        pass

    def router_removed_from_agent(self, context, router_id, host):
        pass

    def router_added_to_agent(self, context, router_ids, host):
        pass

    def routers_updated_on_host(self, context, router_ids, host):
        pass



# Copyright 2013 VMware, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from vmware_nsx._i18n import _


class VcnsException(exceptions.NeutronException):
    pass


class VcnsGeneralException(VcnsException):
    def __init__(self, message):
        self.message = message
        super(VcnsGeneralException, self).__init__()


class VcnsBadRequest(exceptions.BadRequest):
    pass


class VcnsNotFound(exceptions.NotFound):
    message = _('%(resource)s not found: %(msg)s')


class VcnsApiException(VcnsException):
    message = _("An unknown exception %(status)s occurred: %(response)s.")

    def __init__(self, **kwargs):
        super(VcnsApiException, self).__init__(**kwargs)

        self.status = kwargs.get('status')
        self.header = kwargs.get('header')
        self.response = kwargs.get('response')


class ResourceRedirect(VcnsApiException):
    message = _("Resource %(uri)s has been redirected")


class RequestBad(VcnsApiException):
    message = _("Request %(uri)s is Bad, response %(response)s")


class Forbidden(VcnsApiException):
    message = _("Forbidden: %(uri)s")


class ResourceNotFound(VcnsApiException):
    message = _("Resource %(uri)s not found")


class ResourceTimedOut(VcnsApiException):
    message = _("Resource %(uri)s timed out")


class MediaTypeUnsupport(VcnsApiException):
    message = _("Media Type %(uri)s is not supported")


class ServiceUnavailable(VcnsApiException):
    message = _("Service Unavailable: %(uri)s")


class ServiceConflict(VcnsApiException):
    message = _("Concurrent object access error: %(uri)s")


class AlreadyExists(VcnsApiException):
    message = _("Resource %(resource)s already exists")



# Copyright 2013 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

import sys

from oslo_config import cfg

from neutron.common import config

from vmware_nsx._i18n import _
from vmware_nsx.common import config as nsx_config  # noqa
from vmware_nsx.common import nsx_utils
from vmware_nsx.nsxlib import mh as nsxlib

config.setup_logging()


def help(name):
    print("Usage: %s path/to/neutron/plugin/ini/config/file" % name)
    sys.exit(1)


def get_nsx_controllers(cluster):
    return cluster.nsx_controllers


def config_helper(config_entity, cluster):
    try:
        return nsxlib.do_request('GET',
                                 "/ws.v1/%s?fields=uuid" % config_entity,
                                 cluster=cluster).get('results', [])
    except Exception as e:
        msg = (_("Error '%(err)s' when connecting to controller(s): %(ctl)s.")
               % {'err': str(e),
                  'ctl': ', '.join(get_nsx_controllers(cluster))})
        raise Exception(msg)


def get_control_cluster_nodes(cluster):
    return config_helper("control-cluster/node", cluster)


def get_gateway_services(cluster):
    ret_gw_services = {"L2GatewayServiceConfig": [],
                       "L3GatewayServiceConfig": []}
    gw_services = config_helper("gateway-service", cluster)
    for gw_service in gw_services:
        ret_gw_services[gw_service['type']].append(gw_service['uuid'])
    return ret_gw_services


def get_transport_zones(cluster):
    transport_zones = config_helper("transport-zone", cluster)
    return [transport_zone['uuid'] for transport_zone in transport_zones]


def get_transport_nodes(cluster):
    transport_nodes = config_helper("transport-node", cluster)
    return [transport_node['uuid'] for transport_node in transport_nodes]


def is_transport_node_connected(cluster, node_uuid):
    try:
        return nsxlib.do_request('GET',
                                 "/ws.v1/transport-node/%s/status" % node_uuid,
                                 cluster=cluster)['connection']['connected']
    except Exception as e:
        msg = (_("Error '%(err)s' when connecting to controller(s): %(ctl)s.")
               % {'err': str(e),
                  'ctl': ', '.join(get_nsx_controllers(cluster))})
        raise Exception(msg)


def main():
    if len(sys.argv) != 2:
        help(sys.argv[0])
    args = ['--config-file']
    args.append(sys.argv[1])
    config.init(args)
    print("----------------------- Database Options -----------------------")
    print("\tconnection: %s" % cfg.CONF.database.connection)
    print("\tretry_interval: %d" % cfg.CONF.database.retry_interval)
    print("\tmax_retries: %d" % cfg.CONF.database.max_retries)
    print("-----------------------    NSX Options   -----------------------")
    print("\tNSX Generation Timeout %d" % cfg.CONF.NSX.nsx_gen_timeout)
    print("\tNumber of concurrent connections to each controller %d" %
          cfg.CONF.NSX.concurrent_connections)
    print("\tmax_lp_per_bridged_ls: %s" % cfg.CONF.NSX.max_lp_per_bridged_ls)
    print("\tmax_lp_per_overlay_ls: %s" % cfg.CONF.NSX.max_lp_per_overlay_ls)
    print("-----------------------  Cluster Options -----------------------")
    print("\tretries: %s" % cfg.CONF.retries)
    print("\tredirects: %s" % cfg.CONF.redirects)
    print("\thttp_timeout: %s" % cfg.CONF.http_timeout)
    cluster = nsx_utils.create_nsx_cluster(
        cfg.CONF,
        cfg.CONF.NSX.concurrent_connections,
        cfg.CONF.NSX.nsx_gen_timeout)
    nsx_controllers = get_nsx_controllers(cluster)
    num_controllers = len(nsx_controllers)
    print("Number of controllers found: %s" % num_controllers)
    if num_controllers == 0:
        print("You must specify at least one controller!")
        sys.exit(1)

    get_control_cluster_nodes(cluster)
    for controller in nsx_controllers:
        print("\tController endpoint: %s" % controller)
        gateway_services = get_gateway_services(cluster)
        default_gateways = {
            "L2GatewayServiceConfig": cfg.CONF.default_l2_gw_service_uuid,
            "L3GatewayServiceConfig": cfg.CONF.default_l3_gw_service_uuid}
        errors = 0
        for svc_type in default_gateways.keys():
            for uuid in gateway_services[svc_type]:
                print("\t\tGateway(%s) uuid: %s" % (svc_type, uuid))
            if (default_gateways[svc_type] and
                default_gateways[svc_type] not in gateway_services[svc_type]):
                print("\t\t\tError: specified default %s gateway (%s) is "
                      "missing from NSX Gateway Services!" % (
                          svc_type,
                          default_gateways[svc_type]))
                errors += 1
        transport_zones = get_transport_zones(cluster)
        print("\tTransport zones: %s" % transport_zones)
        if cfg.CONF.default_tz_uuid not in transport_zones:
            print("\t\tError: specified default transport zone "
                  "(%s) is missing from NSX transport zones!"
                  % cfg.CONF.default_tz_uuid)
            errors += 1
        transport_nodes = get_transport_nodes(cluster)
        print("\tTransport nodes: %s" % transport_nodes)
        node_errors = []
        for node in transport_nodes:
            if not is_transport_node_connected(cluster, node):
                node_errors.append(node)

    # Use different exit codes, so that we can distinguish
    # between config and runtime errors
    if len(node_errors):
        print("\nThere are one or more transport nodes that are "
              "not connected: %s. Please, revise!" % node_errors)
        sys.exit(10)
    elif errors:
        print("\nThere are %d errors with your configuration. "
              "Please, revise!" % errors)
        sys.exit(12)
    else:
        print("Done.")



# Copyright (c) 2015 VMware, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.services.l2gateway.common import config
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc
from networking_l2gw.services.l2gateway import plugin as core_l2gw_plugin
from networking_l2gw.tests.unit.db import test_l2gw_db
from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import uuidutils

from neutron.tests import base
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context
from neutron_lib import exceptions as n_exc

from vmware_nsx.services.l2gateway.nsx_v3 import driver as nsx_v3_driver
from vmware_nsx.tests.unit.nsx_v3 import test_plugin as test_nsx_v3_plugin
from vmware_nsxlib.tests.unit.v3 import mocks as nsx_v3_mocks
from vmware_nsxlib.v3 import nsx_constants


NSX_V3_PLUGIN_CLASS = ('vmware_nsx.plugins.nsx_v3.plugin.NsxV3Plugin')
NSX_V3_L2GW_DRIVER_CLASS_PATH = ('vmware_nsx.services.l2gateway.'
                                 'nsx_v3.driver.NsxV3Driver')


class TestNsxV3L2GatewayDriver(test_l2gw_db.L2GWTestCase,
                               test_nsx_v3_plugin.NsxV3PluginTestCaseMixin,
                               base.BaseTestCase):

    def setUp(self):
        super(TestNsxV3L2GatewayDriver, self).setUp()
        self.core_plugin = importutils.import_object(NSX_V3_PLUGIN_CLASS)
        self.driver = nsx_v3_driver.NsxV3Driver(mock.MagicMock())
        mock.patch.object(config, 'register_l2gw_opts_helper')
        mock.patch('neutron.services.service_base.load_drivers',
                   return_value=({'dummyprovider': self.driver},
                                 'dummyprovider')).start()
        mock.patch.object(l2gateway_db.L2GatewayMixin, '__init__'),
        mock.patch.object(l2gateway_db, 'subscribe')
        mock.patch('neutron.db.servicetype_db.ServiceTypeManager.get_instance',
                   return_value=mock.MagicMock()).start()
        self.l2gw_plugin = core_l2gw_plugin.L2GatewayPlugin()
        self.context = context.get_admin_context()

    def _get_nw_data(self):
        net_data = super(TestNsxV3L2GatewayDriver, self)._get_nw_data()
        net_data['network']['port_security_enabled'] = True
        return net_data

    def test_nsxl2gw_driver_init(self):
        with mock.patch.object(nsx_v3_driver.NsxV3Driver,
                               'subscribe_callback_notifications') as sub:
            with mock.patch.object(nsx_v3_driver.LOG,
                                   'debug') as debug:
                nsx_v3_driver.NsxV3Driver(mock.MagicMock())
                self.assertTrue(sub.called)
                self.assertTrue(debug.called)

    def test_create_default_l2_gateway(self):
        def_bridge_cluster_name = nsx_v3_mocks.NSX_BRIDGE_CLUSTER_NAME
        cfg.CONF.set_override("default_bridge_cluster",
                              def_bridge_cluster_name,
                              "nsx_v3")
        nsx_v3_driver.NsxV3Driver(mock.MagicMock())
        # fake the callback invoked after init
        registry.publish(resources.PROCESS, events.BEFORE_SPAWN,
                         mock.MagicMock())
        l2gws = self.driver._get_l2_gateways(self.context)
        def_bridge_cluster_id = (
            self.nsxlib.bridge_cluster.get_id_by_name_or_id(
                def_bridge_cluster_name))
        def_l2gw = None
        for l2gw in l2gws:
            for device in l2gw['devices']:
                if device['device_name'] == def_bridge_cluster_id:
                    def_l2gw = l2gw
        self.assertIsNotNone(def_l2gw)
        self.assertTrue(def_l2gw.devices[0].device_name,
                        def_bridge_cluster_id)
        self.assertTrue(def_l2gw.devices[0].interfaces[0].interface_name,
                        'default-bridge-cluster')

    def test_create_duplicate_default_l2_gateway_noop(self):
        def_bridge_cluster_name = nsx_v3_mocks.NSX_BRIDGE_CLUSTER_NAME
        cfg.CONF.set_override("default_bridge_cluster",
                              def_bridge_cluster_name,
                              "nsx_v3")
        for i in range(0, 2):
            nsx_v3_driver.NsxV3Driver(mock.MagicMock())
            # fake the callback invoked after init
            registry.publish(resources.PROCESS, events.BEFORE_SPAWN,
                             mock.MagicMock())
        l2gws = self.driver._get_l2_gateways(self.context)
        # Verify whether only one default L2 gateway is created
        self.assertEqual(1, len(l2gws))

    def test_create_default_l2_gateway_no_bc_uuid_noop(self):
        with mock.patch.object(nsx_v3_driver.NsxV3Driver,
                               'subscribe_callback_notifications'):
            nsx_v3_driver.NsxV3Driver(mock.MagicMock())
            l2gws = self.driver._get_l2_gateways(self.context)
            # Verify no default L2 gateway is created if bridge cluster id is
            # not configured in nsx.ini
            self.assertEqual([], l2gws)

    def test_create_l2_gateway_multiple_devices_fail(self):
        invalid_l2gw_dict = {
            "l2_gateway": {
                "tenant_id": "fake_tenant_id",
                "name": "invalid_l2gw",
                "devices": [{"interfaces":
                            [{"name": "interface1"}],
                            "device_name": "device1"},
                            {"interfaces":
                            [{"name": "interface_2"}],
                            "device_name": "device2"}]}}
        self.assertRaises(n_exc.InvalidInput,
                          self.l2gw_plugin.create_l2_gateway,
                          self.context, invalid_l2gw_dict)

    def test_create_l2_gateway_multiple_interfaces_fail(self):
        invalid_l2gw_dict = {
            "l2_gateway": {
                "tenant_id": "fake_tenant_id",
                "name": "invalid_l2gw",
                "devices": [{"interfaces":
                            [{"name": "interface1"},
                             {"name": "interface2"}],
                            "device_name": "device1"}]}}
        self.assertRaises(n_exc.InvalidInput,
                          self.l2gw_plugin.create_l2_gateway,
                          self.context, invalid_l2gw_dict)

    def test_create_l2_gateway_invalid_device_name_fail(self):
        invalid_l2gw_dict = {
            "l2_gateway": {
                "tenant_id": "fake_tenant_id",
                "name": "invalid_l2gw",
                "devices": [{"interfaces":
                            [{"name": "interface_1"}],
                            "device_name": "device-1"}]}}
        self.assertRaises(n_exc.InvalidInput,
                          self.l2gw_plugin.create_l2_gateway,
                          self.context, invalid_l2gw_dict)

    def test_create_l2_gateway_valid(self):
        bc_uuid = uuidutils.generate_uuid()
        l2gw_data = self._get_l2_gateway_data(name='gw1',
                                              device_name=bc_uuid)
        l2gw = self.l2gw_plugin.create_l2_gateway(self.context, l2gw_data)
        self.assertIsNotNone(l2gw)
        self.assertEqual("gw1", l2gw["name"])
        self.assertEqual("port1",
                         l2gw["devices"][0]["interfaces"][0]["name"])
        self.assertEqual(bc_uuid, l2gw["devices"][0]["device_name"])

    def test_create_l2_gateway_connection(self):
        type(self.driver)._core_plugin = self.core_plugin
        bc_uuid = uuidutils.generate_uuid()
        l2gw_data = self._get_l2_gateway_data(name='def-l2gw',
                                              device_name=bc_uuid)
        l2gw = self._create_l2gateway(l2gw_data)
        net_data = self._get_nw_data()
        net = self.core_plugin.create_network(self.context, net_data)
        l2gw_conn_data = {constants.CONNECTION_RESOURCE_NAME: {
            'l2_gateway_id': l2gw['id'],
            'tenant_id': 'fake_tenant_id',
            'network_id': net['id']}}
        l2gw_conn = self.l2gw_plugin.create_l2_gateway_connection(
            self.context, l2gw_conn_data)
        self.assertIsNotNone(l2gw_conn)
        self.assertEqual(net['id'], l2gw_conn['network_id'])
        self.assertEqual(l2gw['id'], l2gw_conn['l2_gateway_id'])

    def test_delete_l2_gateway_connection(self):
        type(self.driver)._core_plugin = self.core_plugin
        bc_uuid = uuidutils.generate_uuid()
        l2gw_data = self._get_l2_gateway_data(name='def-l2gw',
                                              device_name=bc_uuid)
        l2gw = self._create_l2gateway(l2gw_data)
        net_data = self._get_nw_data()
        net = self.core_plugin.create_network(self.context, net_data)
        l2gw_conn_data = {constants.CONNECTION_RESOURCE_NAME: {
            'l2_gateway_id': l2gw['id'],
            'tenant_id': 'fake_tenant_id',
            'network_id': net['id']}}
        l2gw_conn = self.l2gw_plugin.create_l2_gateway_connection(
            self.context,
            l2gw_conn_data)
        self.l2gw_plugin.delete_l2_gateway_connection(self.context,
                                                      l2gw_conn['id'])
        # Verify that the L2 gateway connection was deleted
        self.assertRaises(l2gw_exc.L2GatewayConnectionNotFound,
                          self.l2gw_plugin.get_l2_gateway_connection,
                          self.context, l2gw_conn['id'])
        ports = self.core_plugin.get_ports(self.context)
        # Verify that the L2 gateway connection port was cleaned up
        self.assertEqual(0, len(ports))

    def test_create_l2_gateway_connection_creates_port(self):
        type(self.driver)._core_plugin = self.core_plugin
        bc_uuid = uuidutils.generate_uuid()
        l2gw_data = self._get_l2_gateway_data(name='def-l2gw',
                                              device_name=bc_uuid)
        l2gw = self._create_l2gateway(l2gw_data)
        net_data = self._get_nw_data()
        net = self.core_plugin.create_network(self.context, net_data)
        l2gw_conn_data = {
            'id': uuidutils.generate_uuid(),
            'l2_gateway_id': l2gw['id'],
            'tenant_id': 'fake_tenant_id',
            'network_id': net['id']}
        self.driver.create_l2_gateway_connection_postcommit(self.context,
                                                            l2gw_conn_data)
        ports = self.core_plugin.get_ports(self.context)
        # Verify that the L2 gateway connection port was created with device
        # owner BRIDGEENDPOINT
        self.assertEqual(1, len(ports))
        port = ports[0]
        self.assertEqual(nsx_constants.BRIDGE_ENDPOINT, port['device_owner'])
        # Verify that the L2 gateway connection port was created with no
        # fixed ips
        self.assertEqual(0, len(port.get('fixed_ips')))

    def test_core_plugin_delete_l2_gateway_connection_port_fail(self):
        type(self.driver)._core_plugin = self.core_plugin
        bc_uuid = uuidutils.generate_uuid()
        l2gw_data = self._get_l2_gateway_data(name='def-l2gw',
                                              device_name=bc_uuid)
        l2gw = self._create_l2gateway(l2gw_data)
        net_data = self._get_nw_data()
        net = self.core_plugin.create_network(self.context, net_data)
        l2gw_conn_data = {
            'id': uuidutils.generate_uuid(),
            'l2_gateway_id': l2gw['id'],
            'tenant_id': 'fake_tenant_id',
            'network_id': net['id']}
        self.driver.create_l2_gateway_connection_postcommit(self.context,
                                                            l2gw_conn_data)
        port = self.core_plugin.get_ports(self.context)[0]
        self.assertRaises(n_exc.ServicePortInUse,
                          self.core_plugin.delete_port,
                          self.context, port['id'])



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from sqlalchemy.orm import exc

from neutron.db import _model_query as model_query
from neutron.db import _resource_extend as resource_extend
from neutron.db import _utils as db_utils
from neutron.db import api as db_api
from neutron.db import models_v2

from neutron_lib.api.definitions import network as net_def
from neutron_lib.api.definitions import port as port_def

from oslo_log import log
from oslo_utils import uuidutils

from vmware_nsx.db import nsx_models
from vmware_nsx.extensions import qos_queue as qos

LOG = log.getLogger(__name__)


@resource_extend.has_resource_extenders
class QoSDbMixin(qos.QueuePluginBase):
    """Mixin class to add queues."""

    def create_qos_queue(self, context, qos_queue):
        q = qos_queue['qos_queue']
        with db_api.context_manager.writer.using(context):
            qos_queue = nsx_models.QoSQueue(
                id=q.get('id', uuidutils.generate_uuid()),
                name=q.get('name'),
                tenant_id=q['tenant_id'],
                default=q.get('default'),
                min=q.get('min'),
                max=q.get('max'),
                qos_marking=q.get('qos_marking'),
                dscp=q.get('dscp'))
            context.session.add(qos_queue)
        return self._make_qos_queue_dict(qos_queue)

    def get_qos_queue(self, context, queue_id, fields=None):
        return self._make_qos_queue_dict(
            self._get_qos_queue(context, queue_id), fields)

    def _get_qos_queue(self, context, queue_id):
        try:
            return model_query.get_by_id(context, nsx_models.QoSQueue,
                                         queue_id)
        except exc.NoResultFound:
            raise qos.QueueNotFound(id=queue_id)

    def get_qos_queues(self, context, filters=None, fields=None, sorts=None,
                       limit=None, marker=None, page_reverse=False):
        marker_obj = db_utils.get_marker_obj(self, context, 'qos_queue',
                                             limit, marker)
        return model_query.get_collection(context, nsx_models.QoSQueue,
                                          self._make_qos_queue_dict,
                                          filters=filters, fields=fields,
                                          sorts=sorts, limit=limit,
                                          marker_obj=marker_obj,
                                          page_reverse=page_reverse)

    def delete_qos_queue(self, context, queue_id):
        with db_api.context_manager.writer.using(context):
            qos_queue = self._get_qos_queue(context, queue_id)
            context.session.delete(qos_queue)

    def _process_port_queue_mapping(self, context, port_data, queue_id):
        port_data[qos.QUEUE] = queue_id
        if not queue_id:
            return
        with db_api.context_manager.writer.using(context):
            context.session.add(nsx_models.PortQueueMapping(
                port_id=port_data['id'],
                                queue_id=queue_id))

    def _get_port_queue_bindings(self, context, filters=None, fields=None):
        return model_query.get_collection(context,
                                          nsx_models.PortQueueMapping,
                                          self._make_port_queue_binding_dict,
                                          filters=filters, fields=fields)

    def _delete_port_queue_mapping(self, context, port_id):
        query = model_query.query_with_hooks(context,
                                             nsx_models.PortQueueMapping)
        try:
            binding = query.filter(
                nsx_models.PortQueueMapping.port_id == port_id).one()
        except exc.NoResultFound:
            # return since this can happen if we are updating a port that
            # did not already have a queue on it. There is no need to check
            # if there is one before deleting if we return here.
            return
        with db_api.context_manager.writer.using(context):
            context.session.delete(binding)

    def _process_network_queue_mapping(self, context, net_data, queue_id):
        net_data[qos.QUEUE] = queue_id
        if not queue_id:
            return
        with db_api.context_manager.writer.using(context):
            context.session.add(
                nsx_models.NetworkQueueMapping(network_id=net_data['id'],
                                               queue_id=queue_id))

    def _get_network_queue_bindings(self, context, filters=None, fields=None):
        return model_query.get_collection(
            context,
            nsx_models.NetworkQueueMapping,
            self._make_network_queue_binding_dict,
            filters=filters, fields=fields)

    def _delete_network_queue_mapping(self, context, network_id):
        query = self._model_query(context, nsx_models.NetworkQueueMapping)
        with db_api.context_manager.writer.using(context):
            binding = query.filter_by(network_id=network_id).first()
            if binding:
                context.session.delete(binding)

    @staticmethod
    @resource_extend.extends([net_def.COLLECTION_NAME])
    @resource_extend.extends([port_def.COLLECTION_NAME])
    def _extend_dict_qos_queue(obj_res, obj_db):
        queue_mapping = obj_db['qos_queue']
        if queue_mapping:
            obj_res[qos.QUEUE] = queue_mapping.get('queue_id')
        return obj_res

    def _make_qos_queue_dict(self, queue, fields=None):
        res = {'id': queue['id'],
               'name': queue.get('name'),
               'default': queue.get('default'),
               'tenant_id': queue['tenant_id'],
               'min': queue.get('min'),
               'max': queue.get('max'),
               'qos_marking': queue.get('qos_marking'),
               'dscp': queue.get('dscp')}
        return db_utils.resource_fields(res, fields)

    def _make_port_queue_binding_dict(self, queue, fields=None):
        res = {'port_id': queue['port_id'],
               'queue_id': queue['queue_id']}
        return db_utils.resource_fields(res, fields)

    def _make_network_queue_binding_dict(self, queue, fields=None):
        res = {'network_id': queue['network_id'],
               'queue_id': queue['queue_id']}
        return db_utils.resource_fields(res, fields)

    def _check_for_queue_and_create(self, context, port):
        """Check for queue and create.

        This function determines if a port should be associated with a
        queue. It works by first querying NetworkQueueMapping to determine
        if the network is associated with a queue. If so, then it queries
        NetworkQueueMapping for all the networks that are associated with
        this queue. Next, it queries against all the ports on these networks
        with the port device_id. Finally it queries PortQueueMapping. If that
        query returns a queue_id that is returned. Otherwise a queue is
        created that is the size of the queue associated with the network and
        that queue_id is returned.

        If the network is not associated with a queue we then query to see
        if there is a default queue in the system. If so, a copy of that is
        created and the queue_id is returned.

        Otherwise None is returned. None is also returned if the port does not
        have a device_id or if the device_owner is network:
        """

        queue_to_create = None
        # If there is no device_id don't create a queue. The queue will be
        # created on update port when the device_id is present. Also don't
        # apply QoS to network ports.
        if (not port.get('device_id') or
            port['device_owner'].startswith('network:')):
            return

        # Check if there is a queue associated with the network
        filters = {'network_id': [port['network_id']]}
        network_queue_id = self._get_network_queue_bindings(
            context, filters, ['queue_id'])
        if network_queue_id:
            # get networks that queue is associated with
            filters = {'queue_id': [network_queue_id[0]['queue_id']]}
            networks_with_same_queue = self._get_network_queue_bindings(
                context, filters)

            # get the ports on these networks with the same_queue and device_id
            filters = {'device_id': [port.get('device_id')],
                       'network_id': [network['network_id'] for
                                      network in networks_with_same_queue]}
            query = model_query.query_with_hooks(context, models_v2.Port.id)
            model_query.apply_filters(query, models_v2.Port,
                                      filters, context)
            ports_ids = [p[0] for p in query]
            if ports_ids:
                # shared queue already exists find the queue id
                queues = self._get_port_queue_bindings(context,
                                                       {'port_id': ports_ids},
                                                       ['queue_id'])
                if queues:
                    return queues[0]['queue_id']

            # get the size of the queue we want to create
            queue_to_create = self.get_qos_queue(
                context, network_queue_id[0]['queue_id'])

        else:
            # check for default queue
            filters = {'default': [True]}
            # context is elevated since default queue is owned by admin
            queue_to_create = self.get_qos_queues(context.elevated(), filters)
            if not queue_to_create:
                return
            queue_to_create = queue_to_create[0]

        # create the queue
        tenant_id = port['tenant_id']
        if port.get(qos.RXTX_FACTOR) and queue_to_create.get('max'):
            queue_to_create['max'] = int(queue_to_create['max'] *
                                         port[qos.RXTX_FACTOR])
        queue = {'qos_queue': {'name': queue_to_create.get('name'),
                               'min': queue_to_create.get('min'),
                               'max': queue_to_create.get('max'),
                               'dscp': queue_to_create.get('dscp'),
                               'qos_marking':
                               queue_to_create.get('qos_marking'),
                               'tenant_id': tenant_id}}
        return self.create_qos_queue(context, queue, False)['id']

    def _validate_qos_queue(self, context, qos_queue):
        if qos_queue.get('default'):
            if context.is_admin:
                if self.get_qos_queues(context, filters={'default': [True]}):
                    raise qos.DefaultQueueAlreadyExists()
            else:
                raise qos.DefaultQueueCreateNotAdmin()
        if qos_queue.get('qos_marking') == 'trusted':
            dscp = qos_queue.pop('dscp')
            if dscp:
                # must raise because a non-zero dscp was provided
                raise qos.QueueInvalidMarking()
            LOG.info("DSCP value (%s) will be ignored with 'trusted' "
                     "marking", dscp)
        max = qos_queue.get('max')
        min = qos_queue.get('min')
        # Max can be None
        if max and min > max:
            raise qos.QueueMinGreaterMax()



# Copyright 2013 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet import greenthread
import mock

from neutron.tests import base
from neutron_lib import context as neutron_context
from oslo_config import cfg
import six

from vmware_nsx.common import exceptions as nsxv_exc
from vmware_nsx.plugins.nsx_v import availability_zones as nsx_az
from vmware_nsx.plugins.nsx_v.vshield.common import (
    constants as vcns_const)
from vmware_nsx.plugins.nsx_v.vshield import edge_appliance_driver as e_drv
from vmware_nsx.plugins.nsx_v.vshield.tasks import (
    constants as ts_const)
from vmware_nsx.plugins.nsx_v.vshield.tasks import tasks as ts
from vmware_nsx.plugins.nsx_v.vshield import vcns_driver
from vmware_nsx.tests import unit as vmware
from vmware_nsx.tests.unit.nsx_v.vshield import fake_vcns

VCNS_CONFIG_FILE = vmware.get_fake_conf("vcns.ini.test")

ts.TaskManager.set_default_interval(100)


class VcnsDriverTaskManagerTestCase(base.BaseTestCase):

    def setUp(self):
        super(VcnsDriverTaskManagerTestCase, self).setUp()
        self.manager = ts.TaskManager()
        self.manager.start(100)

    def tearDown(self):
        self.manager.stop()
        # Task manager should not leave running threads around
        # if _thread is None it means it was killed in stop()
        self.assertIsNone(self.manager._thread)
        super(VcnsDriverTaskManagerTestCase, self).tearDown()

    def _test_task_manager_task_process_state(self, sync_exec=False):
        def _task_failed(task, reason):
            task.userdata['result'] = False
            task.userdata['error'] = reason

        def _check_state(task, exp_state):
            if not task.userdata.get('result', True):
                return False

            state = task.userdata['state']
            if state != exp_state:
                msg = "state %d expect %d" % (
                    state, exp_state)
                _task_failed(task, msg)
                return False

            task.userdata['state'] = state + 1
            return True

        def _exec(task):
            if not _check_state(task, 1):
                return ts_const.TaskStatus.ERROR

            if task.userdata['sync_exec']:
                return ts_const.TaskStatus.COMPLETED
            else:
                return ts_const.TaskStatus.PENDING

        def _status(task):
            if task.userdata['sync_exec']:
                _task_failed(task, "_status callback triggered")

            state = task.userdata['state']
            if state == 3:
                _check_state(task, 3)
                return ts_const.TaskStatus.PENDING
            else:
                _check_state(task, 4)
                return ts_const.TaskStatus.COMPLETED

        def _result(task):
            if task.userdata['sync_exec']:
                exp_state = 3
            else:
                exp_state = 5

            _check_state(task, exp_state)

        def _start_monitor(task):
            _check_state(task, 0)

        def _executed_monitor(task):
            _check_state(task, 2)

        def _result_monitor(task):
            if task.userdata['sync_exec']:
                exp_state = 4
            else:
                exp_state = 6

            if _check_state(task, exp_state):
                task.userdata['result'] = True
            else:
                task.userdata['result'] = False

        userdata = {
            'state': 0,
            'sync_exec': sync_exec
        }
        task = ts.Task('name', 'res', _exec, _status, _result, userdata)
        task.add_start_monitor(_start_monitor)
        task.add_executed_monitor(_executed_monitor)
        task.add_result_monitor(_result_monitor)

        self.manager.add(task)

        task.wait(ts_const.TaskState.RESULT)

        self.assertTrue(userdata['result'])

    def test_task_manager_task_sync_exec_process_state(self):
        self._test_task_manager_task_process_state(sync_exec=True)

    def test_task_manager_task_async_exec_process_state(self):
        self._test_task_manager_task_process_state(sync_exec=False)

    def test_task_manager_task_ordered_process(self):
        def _task_failed(task, reason):
            task.userdata['result'] = False
            task.userdata['error'] = reason

        def _exec(task):
            task.userdata['executed'] = True
            return ts_const.TaskStatus.PENDING

        def _status(task):
            return ts_const.TaskStatus.COMPLETED

        def _result(task):
            next_task = task.userdata.get('next')
            if next_task:
                if next_task.userdata.get('executed'):
                    _task_failed(next_task, "executed premature")
            if task.userdata.get('result', True):
                task.userdata['result'] = True

        tasks = []
        prev = None
        last_task = None
        for i in range(5):
            name = "name-%d" % i
            task = ts.Task(name, 'res', _exec, _status, _result, {})
            tasks.append(task)
            if prev:
                prev.userdata['next'] = task
            prev = task
            last_task = task

        for task in tasks:
            self.manager.add(task)

        last_task.wait(ts_const.TaskState.RESULT)

        for task in tasks:
            self.assertTrue(task.userdata['result'])

    def test_task_manager_task_parallel_process(self):
        tasks = []

        def _exec(task):
            task.userdata['executed'] = True
            return ts_const.TaskStatus.PENDING

        def _status(task):
            for t in tasks:
                if not t.userdata.get('executed'):
                    t.userdata['resut'] = False
            return ts_const.TaskStatus.COMPLETED

        def _result(task):
            if (task.userdata.get('result') is None and
                task.status == ts_const.TaskStatus.COMPLETED):
                task.userdata['result'] = True
            else:
                task.userdata['result'] = False

        for i in range(5):
            name = "name-%d" % i
            res = 'resource-%d' % i
            task = ts.Task(name, res, _exec, _status, _result, {})
            tasks.append(task)
            self.manager.add(task)

        for task in tasks:
            task.wait(ts_const.TaskState.RESULT)
            self.assertTrue(task.userdata['result'])

    def _test_task_manager_stop(self, exec_wait=False, result_wait=False,
                                stop_wait=0):
        def _exec(task):
            if exec_wait:
                greenthread.sleep(0.01)
            return ts_const.TaskStatus.PENDING

        def _status(task):
            greenthread.sleep(0.01)
            return ts_const.TaskStatus.PENDING

        def _result(task):
            if result_wait:
                greenthread.sleep(0)

        manager = ts.TaskManager().start(100)
        manager.stop()
        # Task manager should not leave running threads around
        # if _thread is None it means it was killed in stop()
        self.assertIsNone(manager._thread)
        manager.start(100)

        alltasks = {}
        for i in range(100):
            res = 'res-%d' % i
            tasks = []
            for i in range(100):
                task = ts.Task('name', res, _exec, _status, _result)
                manager.add(task)
                tasks.append(task)
            alltasks[res] = tasks

        greenthread.sleep(stop_wait)
        manager.stop()
        # Task manager should not leave running threads around
        # if _thread is None it means it was killed in stop()
        self.assertIsNone(manager._thread)

        for res, tasks in six.iteritems(alltasks):
            for task in tasks:
                self.assertEqual(ts_const.TaskStatus.ABORT, task.status)

    def test_task_manager_stop_1(self):
        self._test_task_manager_stop(True, True, 0)

    def test_task_manager_stop_2(self):
        self._test_task_manager_stop(True, True, 1)

    def test_task_manager_stop_3(self):
        self._test_task_manager_stop(False, False, 0)

    def test_task_manager_stop_4(self):
        self._test_task_manager_stop(False, False, 1)

    def test_task_pending_task(self):
        def _exec(task):
            task.userdata['executing'] = True
            while not task.userdata['tested']:
                greenthread.sleep(0)
            task.userdata['executing'] = False
            return ts_const.TaskStatus.COMPLETED

        userdata = {
            'executing': False,
            'tested': False
        }
        manager = ts.TaskManager().start(100)
        task = ts.Task('name', 'res', _exec, userdata=userdata)
        manager.add(task)

        while not userdata['executing']:
            greenthread.sleep(0)
        self.assertTrue(manager.has_pending_task())

        userdata['tested'] = True
        while userdata['executing']:
            greenthread.sleep(0)
        self.assertFalse(manager.has_pending_task())


class VcnsDriverTestCase(base.BaseTestCase):

    def vcns_patch(self):
        instance = self.mock_vcns.start()
        instance.return_value.deploy_edge.side_effect = self.fc.deploy_edge
        instance.return_value.get_edge_id.side_effect = self.fc.get_edge_id
        instance.return_value.get_edge_deploy_status.side_effect = (
            self.fc.get_edge_deploy_status)
        instance.return_value.delete_edge.side_effect = self.fc.delete_edge
        instance.return_value.update_interface.side_effect = (
            self.fc.update_interface)
        instance.return_value.get_nat_config.side_effect = (
            self.fc.get_nat_config)
        instance.return_value.update_nat_config.side_effect = (
            self.fc.update_nat_config)
        instance.return_value.delete_nat_rule.side_effect = (
            self.fc.delete_nat_rule)
        instance.return_value.get_edge_status.side_effect = (
            self.fc.get_edge_status)
        instance.return_value.get_edges.side_effect = self.fc.get_edges
        instance.return_value.update_routes.side_effect = (
            self.fc.update_routes)
        instance.return_value.create_lswitch.side_effect = (
            self.fc.create_lswitch)
        instance.return_value.delete_lswitch.side_effect = (
            self.fc.delete_lswitch)

    def setUp(self):
        super(VcnsDriverTestCase, self).setUp()

        self.ctx = neutron_context.get_admin_context()
        self.temp_e_drv_nsxv_db = e_drv.nsxv_db
        e_drv.nsxv_db = mock.MagicMock()
        self.config_parse(args=['--config-file', VCNS_CONFIG_FILE])

        self.fc = fake_vcns.FakeVcns()
        self.mock_vcns = mock.patch(vmware.VCNS_NAME, autospec=True)
        self.vcns_patch()

        self.addCleanup(self.fc.reset_all)

        self.vcns_driver = vcns_driver.VcnsDriver(self)

        self.az = (nsx_az.NsxVAvailabilityZones().
                   get_default_availability_zone())
        self.edge_id = None
        self.result = None

    def tearDown(self):
        e_drv.nsxv_db = self.temp_e_drv_nsxv_db
        self.vcns_driver.task_manager.stop()
        # Task manager should not leave running threads around
        # if _thread is None it means it was killed in stop()
        self.assertIsNone(self.vcns_driver.task_manager._thread)
        super(VcnsDriverTestCase, self).tearDown()

    def complete_edge_creation(
            self, context, edge_id, name, router_id, dist, deploy_successful,
            availability_zone=None, deploy_metadata=False):
        pass

    def _deploy_edge(self):
        self.edge_id = self.vcns_driver.deploy_edge(
            self.ctx, 'router-id', 'myedge', 'internal-network',
            availability_zone=self.az)
        self.assertEqual('edge-1', self.edge_id)

    def test_deploy_edge_with(self):
        self.vcns_driver.deploy_edge(
            self.ctx, 'router-id', 'myedge', 'internal-network',
            availability_zone=self.az)
        status = self.vcns_driver.get_edge_status('edge-1')
        self.assertEqual(vcns_const.RouterStatus.ROUTER_STATUS_ACTIVE, status)

    def test_deploy_edge_fail(self):
        self.vcns_driver.deploy_edge(
            self.ctx, 'router-1', 'myedge', 'internal-network',
            availability_zone=self.az)
        # self.vcns_driver.deploy_edge(
        #     self.ctx, 'router-2', 'myedge', 'internal-network',
        #     availability_zone=self.az)
        self.assertRaises(
            nsxv_exc.NsxPluginException, self.vcns_driver.deploy_edge,
            self.ctx, 'router-2', 'myedge', 'internal-network',
            availability_zone=self.az)

    def test_get_edge_status(self):
        self._deploy_edge()
        status = self.vcns_driver.get_edge_status(self.edge_id)
        self.assertEqual(vcns_const.RouterStatus.ROUTER_STATUS_ACTIVE, status)

    def test_update_nat_rules(self):
        self._deploy_edge()
        snats = [{
            'src': '192.168.1.0/24',
            'translated': '10.0.0.1'
        }, {
            'src': '192.168.2.0/24',
            'translated': '10.0.0.2'
        }, {
            'src': '192.168.3.0/24',
            'translated': '10.0.0.3'
        }
        ]
        dnats = [{
            'dst': '100.0.0.4',
            'translated': '192.168.1.1'
        }, {
            'dst': '100.0.0.5',
            'translated': '192.168.2.1'
        }
        ]

        result = self.vcns_driver.update_nat_rules(self.edge_id, snats, dnats)
        self.assertTrue(result)

        natcfg = self.vcns_driver.get_nat_config(self.edge_id)
        rules = natcfg['rules']['natRulesDtos']
        self.assertEqual(2 * len(dnats) + len(snats), len(rules))
        self.natEquals(rules[0], dnats[0])
        self.natEquals(rules[1], self.snat_for_dnat(dnats[0]))
        self.natEquals(rules[2], dnats[1])
        self.natEquals(rules[3], self.snat_for_dnat(dnats[1]))
        self.natEquals(rules[4], snats[0])
        self.natEquals(rules[5], snats[1])
        self.natEquals(rules[6], snats[2])

    def test_update_nat_rules_for_all_vnics(self):
        self._deploy_edge()
        snats = [{
            'src': '192.168.1.0/24',
            'translated': '10.0.0.1'
        }, {
            'src': '192.168.2.0/24',
            'translated': '10.0.0.2'
        }, {
            'src': '192.168.3.0/24',
            'translated': '10.0.0.3'
        }
        ]
        dnats = [{
            'dst': '100.0.0.4',
            'translated': '192.168.1.1'
        }, {
            'dst': '100.0.0.5',
            'translated': '192.168.2.1'
        }
        ]

        indices = [0, 1, 2, 3]
        result = self.vcns_driver.update_nat_rules(self.edge_id,
                snats, dnats, indices)
        self.assertTrue(result)

        natcfg = self.vcns_driver.get_nat_config(self.edge_id)
        rules = natcfg['rules']['natRulesDtos']

        self.assertEqual(2 * len(indices) * len(dnats)
                + len(indices) * len(snats), len(rules))

        sorted_rules = sorted(rules, key=lambda k: k['vnic'])
        for i in range(0, len(sorted_rules), 7):
            self.natEquals(sorted_rules[i], dnats[0])
            self.natEquals(sorted_rules[i + 1], self.snat_for_dnat(dnats[0]))
            self.natEquals(sorted_rules[i + 2], dnats[1])
            self.natEquals(sorted_rules[i + 3], self.snat_for_dnat(dnats[1]))
            self.natEquals(sorted_rules[i + 4], snats[0])
            self.natEquals(sorted_rules[i + 5], snats[1])
            self.natEquals(sorted_rules[i + 6], snats[2])

    def test_update_nat_rules_for_specific_vnics(self):
        self._deploy_edge()
        snats = [{
            'src': '192.168.1.0/24',
            'translated': '10.0.0.1',
            'vnic_index': 5
        }, {
            'src': '192.168.2.0/24',
            'translated': '10.0.0.2'
        }, {
            'src': '192.168.3.0/24',
            'translated': '10.0.0.3'
        }
        ]
        dnats = [{
            'dst': '100.0.0.4',
            'translated': '192.168.1.1',
            'vnic_index': 2
        }, {
            'dst': '100.0.0.5',
            'translated': '192.168.2.1'
        }
        ]

        result = self.vcns_driver.update_nat_rules(self.edge_id, snats, dnats)
        self.assertTrue(result)

        natcfg = self.vcns_driver.get_nat_config(self.edge_id)

        rules = natcfg['rules']['natRulesDtos']

        self.assertEqual(2 * len(dnats) + len(snats), len(rules))

        self.natEquals(rules[0], dnats[0])
        self.assertEqual(2, rules[0]['vnic'])
        self.natEquals(rules[1], self.snat_for_dnat(dnats[0]))
        self.assertEqual(2, rules[1]['vnic'])
        self.natEquals(rules[2], dnats[1])
        self.assertNotIn('vnic', rules[2])
        self.natEquals(rules[3], self.snat_for_dnat(dnats[1]))
        self.assertNotIn('vnic', rules[3])
        self.natEquals(rules[4], snats[0])
        self.assertEqual(5, rules[4]['vnic'])
        self.natEquals(rules[5], snats[1])
        self.assertNotIn('vnic', rules[5])
        self.natEquals(rules[6], snats[2])
        self.assertNotIn('vnic', rules[6])

    def snat_for_dnat(self, dnat):
        return {
            'src': dnat['translated'],
            'translated': dnat['dst']
        }

    def natEquals(self, rule, exp):
        addr = exp.get('src')
        if not addr:
            addr = exp.get('dst')

        self.assertEqual(addr, rule['originalAddress'])
        self.assertEqual(exp['translated'], rule['translatedAddress'])

    def test_update_routes(self):
        self._deploy_edge()
        routes = [{
            'cidr': '192.168.1.0/24',
            'nexthop': '169.254.2.1'
        }, {
            'cidr': '192.168.2.0/24',
            'nexthop': '169.254.2.1'
        }, {
            'cidr': '192.168.3.0/24',
            'nexthop': '169.254.2.1'
        }
        ]
        result = self.vcns_driver.update_routes(
            self.edge_id, '10.0.0.1', routes)
        self.assertTrue(result)

    def test_update_interface(self):
        self._deploy_edge()
        self.vcns_driver.update_interface(
            'router-id', self.edge_id, vcns_const.EXTERNAL_VNIC_INDEX,
            'network-id', address='100.0.0.3', netmask='255.255.255.0')

    def test_delete_edge(self):
        self._deploy_edge()
        result = self.vcns_driver.delete_edge(
            self.ctx, 'router-id', self.edge_id)
        self.assertTrue(result)

    def test_create_lswitch(self):
        tz_config = [{
            'transport_zone_uuid': 'tz-uuid'
        }]
        lswitch = self.vcns_driver.create_lswitch('lswitch', tz_config)
        self.assertEqual('lswitch', lswitch['display_name'])
        self.assertEqual('LogicalSwitchConfig', lswitch['type'])
        self.assertIn('uuid', lswitch)

    def test_delete_lswitch(self):
        tz_config = {
            'transport_zone_uuid': 'tz-uuid'
        }
        lswitch = self.vcns_driver.create_lswitch('lswitch', tz_config)
        self.vcns_driver.delete_lswitch(lswitch['uuid'])


class VcnsDriverHATestCase(VcnsDriverTestCase):

    def setUp(self):
        # add edge_ha and ha_datastore to the pre-defined configuration
        self._data_store = 'fake-datastore'
        self._ha_data_store = 'fake-datastore-2'
        cfg.CONF.set_override('ha_datastore_id', self._ha_data_store,
                              group="nsxv")
        cfg.CONF.set_override('edge_ha', True, group="nsxv")

        super(VcnsDriverHATestCase, self).setUp()

        self.vcns_driver.vcns.orig_deploy = self.vcns_driver.vcns.deploy_edge
        self.vcns_driver.vcns.deploy_edge = self._fake_deploy_edge

    def _fake_deploy_edge(self, request):
        # validate the appliance structure in the request,
        # and return the regular (fake) response
        found_app = request['appliances']['appliances']
        self.assertEqual(2, len(found_app))
        self.assertEqual(self._data_store, found_app[0]['datastoreId'])
        self.assertEqual(self._ha_data_store, found_app[1]['datastoreId'])
        return self.vcns_driver.vcns.orig_deploy(request)



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from oslo_log import log as logging

from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import requests as ipam_req

from vmware_nsx._i18n import _
from vmware_nsx.services.ipam.common import driver as common
from vmware_nsxlib.v3 import exceptions as nsx_lib_exc
from vmware_nsxlib.v3 import nsx_constants as error

LOG = logging.getLogger(__name__)


class Nsxv3IpamDriver(common.NsxAbstractIpamDriver):
    """IPAM Driver For NSX-V3 networks."""

    def __init__(self, subnetpool, context):
        super(Nsxv3IpamDriver, self).__init__(subnetpool, context)
        self.nsxlib_ipam = self._nsxlib.ip_pool

        # Mark which updates to the pool are supported
        self.support_update_gateway = True
        self.support_update_pools = True

    @property
    def _subnet_class(self):
        return Nsxv3IpamSubnet

    def _get_cidr_from_request(self, subnet_request):
        return "%s/%s" % (subnet_request.subnet_cidr[0],
                          subnet_request.prefixlen)

    def _get_ranges_from_request(self, subnet_request):
        if subnet_request.allocation_pools:
            ranges = [
                {'start': str(pool[0]), 'end': str(pool[-1])}
                for pool in subnet_request.allocation_pools]
        else:
            ranges = []
        return ranges

    def allocate_backend_pool(self, subnet_request):
        """Create a pool on the NSX backend and return its ID"""
        # name/description length on backend is long, so there is no problem
        name = 'subnet_' + subnet_request.subnet_id
        description = 'OS IP pool for subnet ' + subnet_request.subnet_id
        try:
            response = self.nsxlib_ipam.create(
                self._get_cidr_from_request(subnet_request),
                allocation_ranges=self._get_ranges_from_request(
                    subnet_request),
                display_name=name,
                description=description,
                gateway_ip=subnet_request.gateway_ip)
            nsx_pool_id = response['id']
        except Exception as e:
            #TODO(asarfaty): handle specific errors
            msg = _('Failed to create subnet IPAM: %s') % e
            raise ipam_exc.IpamValueInvalid(message=msg)
        return nsx_pool_id

    def delete_backend_pool(self, nsx_pool_id):
        # Because of the delete_subnet flow in the neutron plugin,
        # some ports still hold IPs from this pool.
        # Those ports be deleted shortly after this function.
        # We need to release those IPs before deleting the backed pool,
        # or else it will fail.
        pool_allocations = self.nsxlib_ipam.get_allocations(nsx_pool_id)
        if pool_allocations and pool_allocations.get('result_count'):
            for allocation in pool_allocations.get('results', []):
                ip_addr = allocation.get('allocation_id')
                try:
                    self.nsxlib_ipam.release(nsx_pool_id, ip_addr)
                except Exception as e:
                    LOG.warning("Failed to release ip %(ip)s from pool "
                                "%(pool)s: %(e)s",
                                {'ip': ip_addr, 'pool': nsx_pool_id, 'e': e})
        try:
            self.nsxlib_ipam.delete(nsx_pool_id)
        except Exception as e:
            LOG.error("Failed to delete IPAM from backend: %s", e)
            # Continue anyway, since this subnet was already removed

    def update_backend_pool(self, nsx_pool_id, subnet_request):
        update_args = {
            'cidr': self._get_cidr_from_request(subnet_request),
            'allocation_ranges': self._get_ranges_from_request(subnet_request),
            'gateway_ip': subnet_request.gateway_ip}
        try:
            self.nsxlib_ipam.update(
                nsx_pool_id, **update_args)
        except nsx_lib_exc.ManagerError as e:
            LOG.error("NSX IPAM failed to update pool %(id)s: "
                      " %(e)s; code %(code)s",
                      {'e': e,
                       'id': nsx_pool_id,
                       'code': e.error_code})
            if (e.error_code == error.ERR_CODE_IPAM_RANGE_MODIFY or
                e.error_code == error.ERR_CODE_IPAM_RANGE_DELETE or
                e.error_code == error.ERR_CODE_IPAM_RANGE_SHRUNK):
                # The change is not allowed: already allocated IPs out of
                # the new range
                raise ipam_exc.InvalidSubnetRequest(
                    reason=_("Already allocated IPs outside of the updated "
                             "pools"))
        except Exception as e:
            # unexpected error
            msg = _('Failed to update subnet IPAM: %s') % e
            raise ipam_exc.IpamValueInvalid(message=msg)


class Nsxv3IpamSubnet(common.NsxAbstractIpamSubnet):
    """Manage IP addresses for the NSX V3 IPAM driver."""

    def __init__(self, subnet_id, nsx_pool_id, ctx, tenant_id):
        super(Nsxv3IpamSubnet, self).__init__(
            subnet_id, nsx_pool_id, ctx, tenant_id)
        self.nsxlib_ipam = self._nsxlib.ip_pool

    def backend_allocate(self, address_request):
        try:
            # allocate a specific IP
            if isinstance(address_request, ipam_req.SpecificAddressRequest):
                # This handles both specific and automatic address requests
                ip_address = str(address_request.address)
                # If this is the subnet gateway IP - no need to allocate it
                subnet = self.get_details()
                if str(subnet.gateway_ip) == ip_address:
                    LOG.info("Skip allocation of gateway-ip for pool %s",
                             self._nsx_pool_id)
                    return ip_address
            else:
                # Allocate any free IP
                ip_address = None
            response = self.nsxlib_ipam.allocate(self._nsx_pool_id,
                                                 ip_addr=ip_address)
            ip_address = response['allocation_id']
        except nsx_lib_exc.ManagerError as e:
            LOG.error("NSX IPAM failed to allocate ip %(ip)s of subnet "
                      "%(id)s: %(e)s; code %(code)s",
                      {'e': e,
                       'ip': ip_address,
                       'id': self._subnet_id,
                       'code': e.error_code})
            if e.error_code == error.ERR_CODE_IPAM_POOL_EXHAUSTED:
                # No more IP addresses available on the pool
                raise ipam_exc.IpAddressGenerationFailure(
                    subnet_id=self._subnet_id)
            if e.error_code == error.ERR_CODE_IPAM_SPECIFIC_IP:
                # The NSX backend  does not support allocation of specific IPs
                # prior to version 2.0.
                msg = (_("NSX-V3 IPAM driver does not support allocation of a "
                         "specific ip %s for port") % ip_address)
                raise NotImplementedError(msg)
            if e.error_code == error.ERR_CODE_IPAM_IP_ALLOCATED:
                # This IP is already in use
                raise ipam_exc.IpAddressAlreadyAllocated(
                    ip=ip_address, subnet_id=self._subnet_id)
            if e.error_code == error.ERR_CODE_OBJECT_NOT_FOUND:
                msg = (_("NSX-V3 IPAM failed to allocate: pool %s was not "
                         "found") % self._nsx_pool_id)
                raise ipam_exc.IpamValueInvalid(message=msg)
            else:
                # another backend error
                raise ipam_exc.IPAllocationFailed()
        except Exception as e:
            LOG.error("NSX IPAM failed to allocate ip %(ip)s of subnet "
                      "%(id)s: %(e)s",
                      {'e': e,
                       'ip': ip_address,
                       'id': self._subnet_id})
            # handle unexpected failures
            raise ipam_exc.IPAllocationFailed()
        return ip_address

    def backend_deallocate(self, ip_address):
        # If this is the subnet gateway IP - no need to allocate it
        subnet = self.get_details()
        if str(subnet.gateway_ip) == ip_address:
            LOG.info("Skip deallocation of gateway-ip for pool %s",
                     self._nsx_pool_id)
            return
        try:
            self.nsxlib_ipam.release(self._nsx_pool_id, ip_address)
        except nsx_lib_exc.ManagerError as e:
            # fail silently
            LOG.error("NSX IPAM failed to free ip %(ip)s of subnet "
                      "%(id)s: %(e)s; code %(code)s",
                      {'e': e,
                       'ip': ip_address,
                       'id': self._subnet_id,
                       'code': e.error_code})

    def get_details(self):
        """Return subnet data as a SpecificSubnetRequest"""
        # get the pool from the backend
        try:
            pool_details = self.nsxlib_ipam.get(self._nsx_pool_id)
        except Exception as e:
            msg = _('Failed to get details for nsx pool: %(id)s: '
                    '%(e)s') % {'id': self._nsx_pool_id, 'e': e}
            raise ipam_exc.IpamValueInvalid(message=msg)

        first_range = pool_details.get('subnets', [None])[0]
        if not first_range:
            msg = _('Failed to get details for nsx pool: %(id)s') % {
                'id': self._nsx_pool_id}
            raise ipam_exc.IpamValueInvalid(message=msg)

        cidr = first_range.get('cidr')
        gateway_ip = first_range.get('gateway_ip')
        pools = []
        for subnet in pool_details.get('subnets', []):
            for ip_range in subnet.get('allocation_ranges', []):
                pools.append(netaddr.IPRange(ip_range.get('start'),
                                             ip_range.get('end')))
        return ipam_req.SpecificSubnetRequest(
            self._tenant_id, self._subnet_id,
            cidr, gateway_ip=gateway_ip, allocation_pools=pools)



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Edge size
COMPACT = 'compact'
LARGE = 'large'
XLARGE = 'xlarge'
QUADLARGE = 'quadlarge'


SHARED = "shared"
EXCLUSIVE = "exclusive"

# Edge type
SERVICE_EDGE = 'service'
VDR_EDGE = 'vdr'

# Internal element purpose
INTER_EDGE_PURPOSE = 'inter_edge_net'

# etc
INTERNAL_TENANT_ID = 'metadata_internal_project'

# L2 gateway edge name prefix
L2_GATEWAY_EDGE = 'L2 bridging'

# An artificial limit for router name length - subtract 1 for the - separator
ROUTER_NAME_LENGTH = (78 - 1)

# LoadBalancer Certificate constants
#NOTE(abhiraut): Number of days specify the total number of days for which the
#                certificate will be active. This certificate will expire in
#                10 years. Once the backend API allows creation of certs which
#                do not expire, the following constant should be removed.
CERT_NUMBER_OF_DAYS = 3650
CSR_REQUEST = ("<csr><subject>"
               "<attribute><key>CN</key><value>metadata.nsx.local</value>"
               "</attribute>"
               "<attribute><key>O</key><value>Organization</value></attribute>"
               "<attribute><key>OU</key><value>Unit</value></attribute>"
               "<attribute><key>L</key><value>Locality</value></attribute>"
               "<attribute><key>ST</key><value>State</value></attribute>"
               "<attribute><key>C</key><value>US</value></attribute>"
               "</subject><algorithm>RSA</algorithm><keySize>2048</keySize>"
               "</csr>")

# Reserved IPs that cannot overlap defined subnets
RESERVED_IPS = ["169.254.128.0/17",
                "169.254.1.0/24",
                "169.254.64.192/26"]

# VPNaaS constants
ENCRYPTION_ALGORITHM_MAP = {
    '3des': '3des',
    'aes-128': 'aes',
    'aes-256': 'aes256'
}

PFS_MAP = {
    'group2': 'dh2',
    'group5': 'dh5'
}

TRANSFORM_PROTOCOL_ALLOWED = ('esp',)

ENCAPSULATION_MODE_ALLOWED = ('tunnel',)



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import uuid

import mock
from neutron.extensions import l3
from neutron.extensions import securitygroup as secgrp
from neutron.tests.unit import _test_extension_portbindings as test_bindings
import neutron.tests.unit.db.test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit.extensions import test_extra_dhcp_opt as test_dhcpopts
import neutron.tests.unit.extensions.test_l3 as test_l3_plugin
import neutron.tests.unit.extensions.test_l3_ext_gw_mode as test_ext_gw_mode
import neutron.tests.unit.extensions.test_securitygroup as ext_sg
from neutron.tests.unit import testlib_api
from neutron_lib.api.definitions import dvr as dvr_apidef
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import l3_ext_gw_mode as l3_egm_apidef
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib import constants
from neutron_lib import context
from neutron_lib import exceptions as ntn_exc
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log
from oslo_utils import uuidutils
from sqlalchemy import exc as sql_exc
import webob.exc

from vmware_nsx.api_client import exception as api_exc
from vmware_nsx.api_client import version as ver_module
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import sync
from vmware_nsx.common import utils
from vmware_nsx.db import db as nsx_db
from vmware_nsx.nsxlib import mh as nsxlib
from vmware_nsx.tests import unit as vmware
from vmware_nsx.tests.unit.extensions import test_metadata
from vmware_nsx.tests.unit.nsx_mh.apiclient import fake
from vmware_nsx.tests.unit import test_utils

LOG = log.getLogger(__name__)


class NsxPluginV2TestCase(test_plugin.NeutronDbPluginV2TestCase):

    def _create_network(self, fmt, name, admin_state_up,
                        arg_list=None, providernet_args=None,
                        set_context=False, tenant_id=None,
                        **kwargs):
        tenant_id = tenant_id or self._tenant_id
        data = {'network': {'name': name,
                            'admin_state_up': admin_state_up,
                            'tenant_id': tenant_id}}
        # Fix to allow the router:external attribute and any other
        # attributes containing a colon to be passed with
        # a double underscore instead
        kwargs = dict((k.replace('__', ':'), v) for k, v in kwargs.items())
        if extnet_apidef.EXTERNAL in kwargs:
            arg_list = (extnet_apidef.EXTERNAL, ) + (arg_list or ())

        attrs = kwargs
        if providernet_args:
            attrs.update(providernet_args)
        for arg in (('admin_state_up', 'tenant_id', 'shared') +
                    (arg_list or ())):
            # Arg must be present
            if arg in kwargs:
                data['network'][arg] = kwargs[arg]
        network_req = self.new_create_request('networks', data, fmt)
        if set_context and tenant_id:
            # create a specific auth context for this request
            network_req.environ['neutron.context'] = context.Context(
                '', tenant_id)
        return network_req.get_response(self.api)

    def setUp(self,
              plugin=vmware.PLUGIN_NAME,
              ext_mgr=None,
              service_plugins=None):
        test_utils.override_nsx_ini_test()
        # mock api client
        self.fc = fake.FakeClient(vmware.STUBS_PATH)
        self.mock_nsx = mock.patch(vmware.NSXAPI_NAME, autospec=True)
        self.mock_instance = self.mock_nsx.start()
        # Avoid runs of the synchronizer looping call
        patch_sync = mock.patch.object(sync, '_start_loopingcall')
        patch_sync.start()

        # Emulate tests against NSX 2.x
        self.mock_instance.return_value.get_version.return_value = (
            ver_module.Version("2.9"))
        self.mock_instance.return_value.request.side_effect = (
            self.fc.fake_request)
        super(NsxPluginV2TestCase, self).setUp(plugin=plugin,
                                               ext_mgr=ext_mgr)
        # Newly created port's status is always 'DOWN' till NSX wires them.
        self.port_create_status = constants.PORT_STATUS_DOWN
        cfg.CONF.set_override('metadata_mode', None, 'NSX')
        self.addCleanup(self.fc.reset_all)


class TestBasicGet(test_plugin.TestBasicGet, NsxPluginV2TestCase):
    pass


class TestV2HTTPResponse(test_plugin.TestV2HTTPResponse, NsxPluginV2TestCase):
    pass


class TestPortsV2(NsxPluginV2TestCase,
                  test_plugin.TestPortsV2,
                  test_bindings.PortBindingsTestCase,
                  test_bindings.PortBindingsHostTestCaseMixin,
                  test_bindings.PortBindingsVnicTestCaseMixin):

    VIF_TYPE = portbindings.VIF_TYPE_OVS
    HAS_PORT_FILTER = True

    def _test_exhaust_ports(self, providernet_args=None):
        with self.network(name='testnet',
                          providernet_args=providernet_args,
                          arg_list=(pnet.NETWORK_TYPE,
                                    pnet.PHYSICAL_NETWORK,
                                    pnet.SEGMENTATION_ID)) as net:
            with self.subnet(network=net) as sub:
                with self.port(subnet=sub):
                    # creating another port should see an exception
                    self._create_port('json', net['network']['id'], 400)

    def test_exhaust_ports_overlay_network(self):
        cfg.CONF.set_override('max_lp_per_overlay_ls', 1, group='NSX')
        self._test_exhaust_ports()

    def test_exhaust_ports_bridged_network(self):
        cfg.CONF.set_override('max_lp_per_bridged_ls', 1, group="NSX")
        providernet_args = {pnet.NETWORK_TYPE: 'flat',
                            pnet.PHYSICAL_NETWORK: 'tzuuid'}
        self._test_exhaust_ports(providernet_args=providernet_args)

    def test_update_port_delete_ip(self):
        # This test case overrides the default because the nsx plugin
        # implements port_security/security groups and it is not allowed
        # to remove an ip address from a port unless the security group
        # is first removed.
        with self.subnet() as subnet:
            with self.port(subnet=subnet) as port:
                data = {'port': {'admin_state_up': False,
                                 'fixed_ips': [],
                                 secgrp.SECURITYGROUPS: []}}
                req = self.new_update_request('ports',
                                              data, port['port']['id'])
                res = self.deserialize('json', req.get_response(self.api))
                self.assertEqual(res['port']['admin_state_up'],
                                 data['port']['admin_state_up'])
                self.assertEqual(res['port']['fixed_ips'],
                                 data['port']['fixed_ips'])

    def test_create_port_name_exceeds_40_chars(self):
        name = 'this_is_a_port_whose_name_is_longer_than_40_chars'
        with self.port(name=name) as port:
            # Assert the neutron name is not truncated
            self.assertEqual(name, port['port']['name'])

    def _verify_no_orphan_left(self, net_id):
        # Verify no port exists on net
        # ie: cleanup on db was successful
        query_params = "network_id=%s" % net_id
        self._test_list_resources('port', [],
                                  query_params=query_params)
        # Also verify no orphan port was left on nsx
        # no port should be there at all
        self.assertFalse(self.fc._fake_lswitch_lport_dict)

    def test_create_port_nsx_error_no_orphan_left(self):
        with mock.patch.object(nsxlib.switch, 'create_lport',
                               side_effect=api_exc.NsxApiException):
            with self.network() as net:
                net_id = net['network']['id']
                self._create_port(self.fmt, net_id,
                                  webob.exc.HTTPInternalServerError.code)
                self._verify_no_orphan_left(net_id)

    def test_create_port_neutron_error_no_orphan_left(self):
        with mock.patch.object(nsx_db, 'add_neutron_nsx_port_mapping',
                               side_effect=ntn_exc.NeutronException):
            with self.network() as net:
                net_id = net['network']['id']
                self._create_port(self.fmt, net_id,
                                  webob.exc.HTTPInternalServerError.code)
                self._verify_no_orphan_left(net_id)

    def test_create_port_db_error_no_orphan_left(self):
        db_exception = db_exc.DBError(
            inner_exception=sql_exc.IntegrityError(mock.ANY,
                                                   mock.ANY,
                                                   mock.ANY))
        with mock.patch.object(nsx_db, 'add_neutron_nsx_port_mapping',
                               side_effect=db_exception):
            with self.network() as net:
                with self.port(device_owner=constants.DEVICE_OWNER_DHCP):
                    self._verify_no_orphan_left(net['network']['id'])

    def test_create_port_maintenance_returns_503(self):
        with self.network() as net:
            with mock.patch.object(nsxlib, 'do_request',
                                   side_effect=nsx_exc.MaintenanceInProgress):
                data = {'port': {'network_id': net['network']['id'],
                                 'admin_state_up': False,
                                 'fixed_ips': [],
                                 'tenant_id': self._tenant_id}}
                plugin = directory.get_plugin()
                with mock.patch.object(plugin, 'get_network',
                                       return_value=net['network']):
                    port_req = self.new_create_request('ports', data, self.fmt)
                    res = port_req.get_response(self.api)
                    self.assertEqual(webob.exc.HTTPServiceUnavailable.code,
                                     res.status_int)


class TestNetworksV2(test_plugin.TestNetworksV2, NsxPluginV2TestCase):

    def test_create_network_vlan_transparent(self):
        self.skipTest("Currently no support in plugin for this")

    def _test_create_bridge_network(self, vlan_id=0):
        net_type = 'vlan' if vlan_id else 'flat'
        name = 'bridge_net'
        expected = [('subnets', []), ('name', name), ('admin_state_up', True),
                    ('status', 'ACTIVE'), ('shared', False),
                    (pnet.NETWORK_TYPE, net_type),
                    (pnet.PHYSICAL_NETWORK, 'tzuuid'),
                    (pnet.SEGMENTATION_ID, vlan_id)]
        providernet_args = {pnet.NETWORK_TYPE: net_type,
                            pnet.PHYSICAL_NETWORK: 'tzuuid'}
        if vlan_id:
            providernet_args[pnet.SEGMENTATION_ID] = vlan_id
        with self.network(name=name,
                          providernet_args=providernet_args,
                          arg_list=(pnet.NETWORK_TYPE,
                                    pnet.PHYSICAL_NETWORK,
                                    pnet.SEGMENTATION_ID)) as net:
            for k, v in expected:
                self.assertEqual(net['network'][k], v)

    def test_create_bridge_network(self):
        self._test_create_bridge_network()

    def test_create_bridge_vlan_network(self):
        self._test_create_bridge_network(vlan_id=123)

    def test_create_bridge_vlan_network_outofrange_returns_400(self):
        with testlib_api.ExpectedException(
                webob.exc.HTTPClientError) as ctx_manager:
            self._test_create_bridge_network(vlan_id=5000)
        self.assertEqual(ctx_manager.exception.code, 400)

    def test_create_l3_ext_network_fails_if_not_external(self):
        net_type = 'l3_ext'
        name = 'l3_ext_net'
        providernet_args = {pnet.NETWORK_TYPE: net_type,
                            pnet.PHYSICAL_NETWORK: 'l3gwuuid',
                            pnet.SEGMENTATION_ID: 123}
        with testlib_api.ExpectedException(
                webob.exc.HTTPClientError) as ctx_manager:
            with self.network(name=name,
                              providernet_args=providernet_args,
                              arg_list=(pnet.NETWORK_TYPE,
                                        pnet.PHYSICAL_NETWORK,
                                        pnet.SEGMENTATION_ID)):
                pass
        self.assertEqual(ctx_manager.exception.code,
                         webob.exc.HTTPBadRequest.code)

    def test_list_networks_filter_by_id(self):
        # We add this unit test to cover some logic specific to the
        # nsx plugin
        with self.network(name='net1') as net1:
            with self.network(name='net2') as net2:
                query_params = 'id=%s' % net1['network']['id']
                self._test_list_resources('network', [net1],
                                          query_params=query_params)
                query_params += '&id=%s' % net2['network']['id']
                self._test_list_resources('network', [net1, net2],
                                          query_params=query_params)

    def test_delete_network_after_removing_subet(self):
        gateway_ip = '10.0.0.1'
        cidr = '10.0.0.0/24'
        fmt = 'json'
        # Create new network
        res = self._create_network(fmt=fmt, name='net',
                                   admin_state_up=True)
        network = self.deserialize(fmt, res)
        subnet = self._make_subnet(fmt, network, gateway_ip,
                                   cidr, ip_version=4)
        req = self.new_delete_request('subnets', subnet['subnet']['id'])
        sub_del_res = req.get_response(self.api)
        self.assertEqual(sub_del_res.status_int, 204)
        req = self.new_delete_request('networks', network['network']['id'])
        net_del_res = req.get_response(self.api)
        self.assertEqual(net_del_res.status_int, 204)

    def test_list_networks_with_shared(self):
        with self.network(name='net1'):
            with self.network(name='net2', shared=True):
                req = self.new_list_request('networks')
                res = self.deserialize('json', req.get_response(self.api))
                self.assertEqual(len(res['networks']), 2)
                req_2 = self.new_list_request('networks')
                req_2.environ['neutron.context'] = context.Context('',
                                                                   'somebody')
                res = self.deserialize('json', req_2.get_response(self.api))
                # tenant must see a single network
                self.assertEqual(len(res['networks']), 1)

    def test_create_network_name_exceeds_40_chars(self):
        name = 'this_is_a_network_whose_name_is_longer_than_40_chars'
        with self.network(name=name) as net:
            # Assert neutron name is not truncated
            self.assertEqual(net['network']['name'], name)

    def test_create_network_maintenance_returns_503(self):
        data = {'network': {'name': 'foo',
                            'admin_state_up': True,
                            'tenant_id': self._tenant_id}}
        with mock.patch.object(nsxlib, 'do_request',
                               side_effect=nsx_exc.MaintenanceInProgress):
            net_req = self.new_create_request('networks', data, self.fmt)
            res = net_req.get_response(self.api)
            self.assertEqual(webob.exc.HTTPServiceUnavailable.code,
                             res.status_int)

    def test_update_network_with_admin_false(self):
        data = {'network': {'admin_state_up': False}}
        with self.network() as net:
            plugin = directory.get_plugin()
            self.assertRaises(NotImplementedError,
                              plugin.update_network,
                              context.get_admin_context(),
                              net['network']['id'], data)

    def test_update_network_with_name_calls_nsx(self):
        with mock.patch.object(
            nsxlib.switch, 'update_lswitch') as update_lswitch_mock:
            # don't worry about deleting this network, do not use
            # context manager
            ctx = context.get_admin_context()
            # Because of commit 79c9712 a tenant must be specified otherwise
            # the unit test will fail
            ctx.tenant_id = 'whatever'
            plugin = directory.get_plugin()
            net = plugin.create_network(
                ctx, {'network': {'name': 'xxx',
                                  'admin_state_up': True,
                                  'shared': False,
                                  'tenant_id': ctx.tenant_id,
                                  'port_security_enabled': True}})
            plugin.update_network(ctx, net['id'],
                                  {'network': {'name': 'yyy'}})
        update_lswitch_mock.assert_called_once_with(
            mock.ANY, mock.ANY, 'yyy')


class SecurityGroupsTestCase(ext_sg.SecurityGroupDBTestCase):

    def setUp(self):
        test_utils.override_nsx_ini_test()
        # mock nsx api client
        self.fc = fake.FakeClient(vmware.STUBS_PATH)
        self.mock_nsx = mock.patch(vmware.NSXAPI_NAME, autospec=True)
        instance = self.mock_nsx.start()
        instance.return_value.login.return_value = "the_cookie"
        # Avoid runs of the synchronizer looping call
        patch_sync = mock.patch.object(sync, '_start_loopingcall')
        patch_sync.start()

        instance.return_value.request.side_effect = self.fc.fake_request
        super(SecurityGroupsTestCase, self).setUp(vmware.PLUGIN_NAME)
        self.plugin = directory.get_plugin()


class TestSecurityGroup(ext_sg.TestSecurityGroups, SecurityGroupsTestCase):

    def test_create_security_group_name_exceeds_40_chars(self):
        name = 'this_is_a_secgroup_whose_name_is_longer_than_40_chars'
        with self.security_group(name=name) as sg:
            # Assert Neutron name is not truncated
            self.assertEqual(sg['security_group']['name'], name)

    def test_create_security_group_rule_bad_input(self):
        name = 'foo security group'
        description = 'foo description'
        with self.security_group(name, description) as sg:
            security_group_id = sg['security_group']['id']
            protocol = 200
            min_range = 32
            max_range = 4343
            rule = self._build_security_group_rule(
                security_group_id, 'ingress', protocol,
                min_range, max_range)
            res = self._create_security_group_rule(self.fmt, rule)
            self.deserialize(self.fmt, res)
            self.assertEqual(res.status_int, 400)

    def test_skip_duplicate_default_sg_error(self):
        num_called = [0]
        original_func = self.plugin.create_security_group

        def side_effect(context, security_group, default_sg):
            # can't always raise, or create_security_group will hang
            self.assertTrue(default_sg)
            self.assertTrue(num_called[0] < 2)
            num_called[0] += 1
            ret = original_func(context, security_group, default_sg)
            if num_called[0] == 1:
                return ret
            # make another call to cause an exception.
            # NOTE(yamamoto): raising the exception by ourselves
            # doesn't update the session state appropriately.
            self.assertRaises(db_exc.DBDuplicateEntry(),
                              original_func, context, security_group,
                              default_sg)

        with mock.patch.object(self.plugin,
                               'create_security_group',
                               side_effect=side_effect):
            self.plugin.create_network(
                context.get_admin_context(),
                {'network': {'name': 'foo',
                             'admin_state_up': True,
                             'shared': False,
                             'tenant_id': 'bar',
                             'port_security_enabled': True}})

    def test_create_security_group_rule_icmpv6_legacy_protocol_name(self):
        self.skipTest('not supported')


class TestL3ExtensionManager(object):

    def get_resources(self):
        # Simulate extension of L3 attribute map
        l3.L3().update_attributes_map(
            l3_egm_apidef.RESOURCE_ATTRIBUTE_MAP)
        l3.L3().update_attributes_map(
            dvr_apidef.RESOURCE_ATTRIBUTE_MAP)
        return l3.L3.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class TestL3SecGrpExtensionManager(TestL3ExtensionManager):
    """A fake extension manager for L3 and Security Group extensions.

    Includes also NSX specific L3 attributes.
    """

    def get_resources(self):
        resources = super(TestL3SecGrpExtensionManager,
                          self).get_resources()
        resources.extend(secgrp.Securitygroup.get_resources())
        return resources


class L3NatTest(test_l3_plugin.L3BaseForIntTests, NsxPluginV2TestCase):

    def setUp(self, plugin=vmware.PLUGIN_NAME, ext_mgr=None,
              service_plugins=None):
        cfg.CONF.set_override('api_extensions_path', vmware.NSXEXT_PATH)
        ext_mgr = ext_mgr or TestL3ExtensionManager()
        super(L3NatTest, self).setUp(
            plugin=plugin, ext_mgr=ext_mgr, service_plugins=service_plugins)
        plugin_instance = directory.get_plugin()
        self._plugin_name = "%s.%s" % (
            plugin_instance.__module__,
            plugin_instance.__class__.__name__)
        self._plugin_class = plugin_instance.__class__

    def _create_l3_ext_network(self, vlan_id=None):
        name = 'l3_ext_net'
        net_type = utils.NetworkTypes.L3_EXT
        providernet_args = {pnet.NETWORK_TYPE: net_type,
                            pnet.PHYSICAL_NETWORK: 'l3_gw_uuid'}
        if vlan_id:
            providernet_args[pnet.SEGMENTATION_ID] = vlan_id
        return self.network(name=name,
                            router__external=True,
                            providernet_args=providernet_args,
                            arg_list=(pnet.NETWORK_TYPE,
                                      pnet.PHYSICAL_NETWORK,
                                      pnet.SEGMENTATION_ID))


class TestL3NatTestCase(L3NatTest,
                        test_l3_plugin.L3NatDBIntTestCase,
                        NsxPluginV2TestCase,
                        test_metadata.MetaDataTestCase):

    def _test_create_l3_ext_network(self, vlan_id=0):
        name = 'l3_ext_net'
        net_type = utils.NetworkTypes.L3_EXT
        expected = [('subnets', []), ('name', name), ('admin_state_up', True),
                    ('status', 'ACTIVE'), ('shared', False),
                    (extnet_apidef.EXTERNAL, True),
                    (pnet.NETWORK_TYPE, net_type),
                    (pnet.PHYSICAL_NETWORK, 'l3_gw_uuid'),
                    (pnet.SEGMENTATION_ID, vlan_id)]
        with self._create_l3_ext_network(vlan_id) as net:
            for k, v in expected:
                self.assertEqual(net['network'][k], v)

    def _nsx_validate_ext_gw(self, router_id, l3_gw_uuid, vlan_id):
        """Verify data on fake NSX API client in order to validate
        plugin did set them properly
        """
        # First find the NSX router ID
        ctx = context.get_admin_context()
        nsx_router_id = nsx_db.get_nsx_router_id(ctx.session, router_id)
        ports = [port for port in self.fc._fake_lrouter_lport_dict.values()
                 if (port['lr_uuid'] == nsx_router_id and
                     port['att_type'] == "L3GatewayAttachment")]
        self.assertEqual(len(ports), 1)
        self.assertEqual(ports[0]['attachment_gwsvc_uuid'], l3_gw_uuid)
        self.assertEqual(ports[0].get('vlan_id'), vlan_id)

    def test_create_l3_ext_network_without_vlan(self):
        self._test_create_l3_ext_network()

    def _test_router_create_with_gwinfo_and_l3_ext_net(self, vlan_id=None,
                                                       validate_ext_gw=True):
        with self._create_l3_ext_network(vlan_id) as net:
            with self.subnet(network=net) as s:
                data = {'router': {'tenant_id': 'whatever'}}
                data['router']['name'] = 'router1'
                data['router']['external_gateway_info'] = {
                    'network_id': s['subnet']['network_id']}
                router_req = self.new_create_request('routers', data,
                                                     self.fmt)
                res = router_req.get_response(self.ext_api)
                router = self.deserialize(self.fmt, res)
                self.assertEqual(
                    s['subnet']['network_id'],
                    (router['router']['external_gateway_info']
                     ['network_id']))
                if validate_ext_gw:
                    self._nsx_validate_ext_gw(router['router']['id'],
                                              'l3_gw_uuid', vlan_id)

    def test_router_create_with_gwinfo_and_l3_ext_net(self):
        self._test_router_create_with_gwinfo_and_l3_ext_net()

    def test_router_create_with_gwinfo_and_l3_ext_net_with_vlan(self):
        self._test_router_create_with_gwinfo_and_l3_ext_net(444)

    def _test_router_create_with_distributed(self, dist_input, dist_expected,
                                             version='3.1', return_code=201):
        self.mock_instance.return_value.get_version.return_value = (
            ver_module.Version(version))

        data = {'tenant_id': 'whatever'}
        data['name'] = 'router1'
        data['distributed'] = dist_input
        router_req = self.new_create_request(
            'routers', {'router': data}, self.fmt)
        res = router_req.get_response(self.ext_api)
        self.assertEqual(return_code, res.status_int)
        if res.status_int == 201:
            router = self.deserialize(self.fmt, res)
            self.assertIn('distributed', router['router'])
            self.assertEqual(dist_expected,
                             router['router']['distributed'])

    def test_router_create_distributed_with_3_1(self):
        self._test_router_create_with_distributed(True, True)

    def test_router_create_distributed_with_new_nsx_versions(self):
        with mock.patch.object(nsxlib.router, 'create_explicit_route_lrouter'):
            self._test_router_create_with_distributed(True, True, '3.2')
            self._test_router_create_with_distributed(True, True, '4.0')
            self._test_router_create_with_distributed(True, True, '4.1')

    def test_router_create_not_distributed(self):
        self._test_router_create_with_distributed(False, False)

    def test_router_create_distributed_unspecified(self):
        self._test_router_create_with_distributed(None, False)

    def test_router_create_distributed_returns_400(self):
        self._test_router_create_with_distributed(True, None, '3.0', 400)

    def test_router_create_on_obsolete_platform(self):

        def obsolete_response(*args, **kwargs):
            response = (nsxlib.router.
                        _create_implicit_routing_lrouter(*args, **kwargs))
            response.pop('distributed')
            return response

        with mock.patch.object(
            nsxlib.router, 'create_lrouter', new=obsolete_response):
            self._test_router_create_with_distributed(None, False, '2.2')

    def _create_router_with_gw_info_for_test(self, subnet):
        data = {'router': {'tenant_id': 'whatever',
                           'name': 'router1',
                           'external_gateway_info':
                           {'network_id': subnet['subnet']['network_id']}}}
        router_req = self.new_create_request(
            'routers', data, self.fmt)
        return router_req.get_response(self.ext_api)

    def test_router_create_nsx_error_returns_500(self, vlan_id=None):
        with mock.patch.object(nsxlib.router,
                               'create_router_lport',
                               side_effect=api_exc.NsxApiException):
            with self._create_l3_ext_network(vlan_id) as net:
                with self.subnet(network=net) as s:
                    res = self._create_router_with_gw_info_for_test(s)
                    self.assertEqual(
                        webob.exc.HTTPInternalServerError.code,
                        res.status_int)

    def test_router_add_gateway_invalid_network_returns_404(self):
        # NOTE(salv-orlando): This unit test has been overridden
        # as the nsx plugin support the ext_gw_mode extension
        # which mandates an uuid for the external network identifier
        with self.router() as r:
            self._add_external_gateway_to_router(
                r['router']['id'],
                uuidutils.generate_uuid(),
                expected_code=webob.exc.HTTPNotFound.code)

    def _verify_router_rollback(self):
        # Check that nothing is left on DB
        # TODO(salv-orlando): Verify whehter this is thread-safe
        # w.r.t. sqllite and parallel testing
        self._test_list_resources('router', [])
        # Check that router is not in NSX
        self.assertFalse(self.fc._fake_lrouter_dict)

    # TODO(asarfaty): make this test pass with the new enginefacade
    def skip_test_router_create_with_gw_info_neutron_fail_does_rollback(self):
        # Simulate get subnet error while building list of ips with prefix
        with mock.patch.object(self._plugin_class,
                               '_build_ip_address_list',
                               side_effect=ntn_exc.SubnetNotFound(
                                   subnet_id='xxx')):
            with self._create_l3_ext_network() as net:
                with self.subnet(network=net) as s:
                    res = self._create_router_with_gw_info_for_test(s)
                    self.assertEqual(
                        webob.exc.HTTPNotFound.code,
                        res.status_int)
                    self._verify_router_rollback()

    def test_router_create_with_gw_info_nsx_fail_does_rollback(self):
        # Simulate error while fetching nsx router gw port
        with mock.patch.object(self._plugin_class,
                               '_find_router_gw_port',
                               side_effect=api_exc.NsxApiException):
            with self._create_l3_ext_network() as net:
                with self.subnet(network=net) as s:
                    res = self._create_router_with_gw_info_for_test(s)
                    self.assertEqual(
                        webob.exc.HTTPInternalServerError.code,
                        res.status_int)
                    self._verify_router_rollback()

    def _test_router_update_gateway_on_l3_ext_net(self, vlan_id=None,
                                                  validate_ext_gw=True):
        with self.router() as r:
            with self.subnet() as s1:
                with self._create_l3_ext_network(vlan_id) as net:
                    with self.subnet(network=net) as s2:
                        self._set_net_external(s1['subnet']['network_id'])
                        try:
                            self._add_external_gateway_to_router(
                                r['router']['id'],
                                s1['subnet']['network_id'])
                            body = self._show('routers', r['router']['id'])
                            net_id = (body['router']
                                      ['external_gateway_info']['network_id'])
                            self.assertEqual(net_id,
                                             s1['subnet']['network_id'])
                            # Plug network with external mapping
                            self._set_net_external(s2['subnet']['network_id'])
                            self._add_external_gateway_to_router(
                                r['router']['id'],
                                s2['subnet']['network_id'])
                            body = self._show('routers', r['router']['id'])
                            net_id = (body['router']
                                      ['external_gateway_info']['network_id'])
                            self.assertEqual(net_id,
                                             s2['subnet']['network_id'])
                            if validate_ext_gw:
                                self._nsx_validate_ext_gw(
                                    body['router']['id'],
                                    'l3_gw_uuid', vlan_id)
                        finally:
                            # Cleanup
                            self._remove_external_gateway_from_router(
                                r['router']['id'],
                                s2['subnet']['network_id'])

    def test_router_update_gateway_on_l3_ext_net(self):
        self._test_router_update_gateway_on_l3_ext_net()

    def test_router_update_gateway_on_l3_ext_net_with_vlan(self):
        self._test_router_update_gateway_on_l3_ext_net(444)

    def test_router_list_by_tenant_id(self):
        with self.router(), self.router():
            with self.router(tenant_id='custom') as router1:
                self._test_list_resources('router', [router1],
                                          query_params="tenant_id=custom")

    def test_create_l3_ext_network_with_vlan(self):
        self._test_create_l3_ext_network(666)

    def test_floatingip_with_assoc_fails(self):
        self._test_floatingip_with_assoc_fails(
            "%s.%s" % (self._plugin_name, "_update_fip_assoc"))

    def test_floatingip_with_invalid_create_port(self):
        self._test_floatingip_with_invalid_create_port(self._plugin_name)

    def test_create_router_name_exceeds_40_chars(self):
        name = 'this_is_a_router_whose_name_is_longer_than_40_chars'
        with self.router(name=name) as rtr:
            # Assert Neutron name is not truncated
            self.assertEqual(rtr['router']['name'], name)

    def test_router_add_interface_port(self):
        orig_update_port = self.plugin.update_port
        with self.router() as r, (
            self.port()) as p, (
                mock.patch.object(self.plugin, 'update_port')) as update_port:
            update_port.side_effect = orig_update_port
            body = self._router_interface_action('add',
                                                 r['router']['id'],
                                                 None,
                                                 p['port']['id'])
            self.assertIn('port_id', body)
            self.assertEqual(p['port']['id'], body['port_id'])
            expected_port_update = {'port_security_enabled': False,
                                    'security_groups': []}
            update_port.assert_any_call(
                mock.ANY, p['port']['id'], {'port': expected_port_update})
            # fetch port and confirm device_id
            body = self._show('ports', p['port']['id'])
            self.assertEqual(r['router']['id'], body['port']['device_id'])

            # clean-up
            self._router_interface_action('remove',
                                          r['router']['id'],
                                          None,
                                          p['port']['id'])

    def _test_floatingip_update(self, expected_status):
        super(TestL3NatTestCase, self).test_floatingip_update(
            expected_status)

    def test_floatingip_update(self):
        self._test_floatingip_update(constants.FLOATINGIP_STATUS_DOWN)

    def test_floatingip_disassociate(self):
        with self.port() as p:
            private_sub = {'subnet': {'id':
                                      p['port']['fixed_ips'][0]['subnet_id']}}
            plugin = directory.get_plugin()
            with mock.patch.object(plugin, 'notify_routers_updated') as notify:
                with self.floatingip_no_assoc(private_sub) as fip:
                    port_id = p['port']['id']
                    body = self._update('floatingips', fip['floatingip']['id'],
                                        {'floatingip': {'port_id': port_id}})
                    self.assertEqual(body['floatingip']['port_id'], port_id)
                    # Floating IP status should be active
                    self.assertEqual(constants.FLOATINGIP_STATUS_ACTIVE,
                                     body['floatingip']['status'])
                    # Disassociate
                    body = self._update('floatingips', fip['floatingip']['id'],
                                        {'floatingip': {'port_id': None}})
                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertIsNone(body['floatingip']['port_id'])
                    self.assertIsNone(body['floatingip']['fixed_ip_address'])
                    # Floating IP status should be down
                    self.assertEqual(constants.FLOATINGIP_STATUS_DOWN,
                                     body['floatingip']['status'])

                # check that notification was not requested
                self.assertFalse(notify.called)

    def test_create_router_maintenance_returns_503(self):
        with self._create_l3_ext_network() as net:
            with self.subnet(network=net) as s:
                with mock.patch.object(
                    nsxlib,
                    'do_request',
                    side_effect=nsx_exc.MaintenanceInProgress):
                    data = {'router': {'tenant_id': 'whatever'}}
                    data['router']['name'] = 'router1'
                    data['router']['external_gateway_info'] = {
                        'network_id': s['subnet']['network_id']}
                    router_req = self.new_create_request(
                        'routers', data, self.fmt)
                    res = router_req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPServiceUnavailable.code,
                                     res.status_int)

    def test_router_add_interface_port_removes_security_group(self):
        with self.router() as r:
            with self.port(do_delete=False) as p:
                body = self._router_interface_action('add',
                                                     r['router']['id'],
                                                     None,
                                                     p['port']['id'])
                self.assertIn('port_id', body)
                self.assertEqual(body['port_id'], p['port']['id'])

                # fetch port and confirm no security-group on it.
                body = self._show('ports', p['port']['id'])
                self.assertEqual(body['port']['security_groups'], [])
                self.assertFalse(body['port']['port_security_enabled'])
                # clean-up
                self._router_interface_action('remove',
                                              r['router']['id'],
                                              None,
                                              p['port']['id'])

    def test_update_subnet_gateway_for_external_net(self):
        plugin = directory.get_plugin()
        port_mock = {'uuid': uuidutils.generate_uuid()}
        with mock.patch.object(plugin, '_find_router_gw_port',
                               return_value=port_mock):
            super(TestL3NatTestCase,
                  self).test_update_subnet_gateway_for_external_net()

    def test_floatingip_update_to_same_port_id_twice(self):
        self.skipTest('Plugin changes floating port status')

    def test_floating_port_status_not_applicable(self):
        self.skipTest('Plugin changes floating port status')

    def test_create_router_gateway_fails(self):
        self.skipTest('not supported')

    def test_first_floatingip_associate_notification(self):
        self.skipTest('not supported')

    def test_floatingip_disassociate_notification(self):
        self.skipTest('not supported')

    def test_metadata_network_with_update_subnet_dhcp_enable(self):
        self.skipTest('not supported')

    def test_metadata_network_with_update_subnet_dhcp_disable(self):
        self.skipTest('not supported')

    def test_floatingip_via_router_interface_returns_404(self):
        self.skipTest('not supported')

    def test_floatingip_via_router_interface_returns_201(self):
        self.skipTest('not supported')

    def test_floatingip_update_subnet_gateway_disabled(self):
        self.skipTest('not supported')


class ExtGwModeTestCase(NsxPluginV2TestCase,
                        test_ext_gw_mode.ExtGwModeIntTestCase):
    def test_router_gateway_set_fail_after_port_create(self):
        self.skipTest("TBD")


class NeutronNsxOutOfSync(NsxPluginV2TestCase,
                          test_l3_plugin.L3NatTestCaseMixin,
                          ext_sg.SecurityGroupsTestCase):

    def setUp(self):
        super(NeutronNsxOutOfSync, self).setUp(
            ext_mgr=TestL3SecGrpExtensionManager())

    def test_delete_network_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        self.fc._fake_lswitch_dict.clear()
        req = self.new_delete_request('networks', net1['network']['id'])
        res = req.get_response(self.api)
        self.assertEqual(res.status_int, 204)

    def test_show_network_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net = self.deserialize('json', res)
        self.fc._fake_lswitch_dict.clear()
        req = self.new_show_request('networks', net['network']['id'],
                                    fields=['id', 'status'])
        net = self.deserialize('json', req.get_response(self.api))
        self.assertEqual(net['network']['status'],
                         constants.NET_STATUS_ERROR)

    def test_delete_port_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        res = self._create_port('json', net1['network']['id'])
        port = self.deserialize('json', res)
        self.fc._fake_lswitch_lport_dict.clear()
        req = self.new_delete_request('ports', port['port']['id'])
        res = req.get_response(self.api)
        self.assertEqual(res.status_int, 204)

    def test_show_port_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        res = self._create_port('json', net1['network']['id'])
        port = self.deserialize('json', res)
        self.fc._fake_lswitch_lport_dict.clear()
        self.fc._fake_lswitch_lportstatus_dict.clear()
        req = self.new_show_request('ports', port['port']['id'],
                                    fields=['id', 'status'])
        net = self.deserialize('json', req.get_response(self.api))
        self.assertEqual(net['port']['status'],
                         constants.PORT_STATUS_ERROR)

    def test_create_port_on_network_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        self.fc._fake_lswitch_dict.clear()
        res = self._create_port('json', net1['network']['id'])
        port = self.deserialize('json', res)
        self.assertEqual(port['port']['status'], constants.PORT_STATUS_ERROR)

    def test_update_port_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        res = self._create_port('json', net1['network']['id'])
        port = self.deserialize('json', res)
        self.fc._fake_lswitch_lport_dict.clear()
        data = {'port': {'name': 'error_port'}}
        req = self.new_update_request('ports', data, port['port']['id'])
        port = self.deserialize('json', req.get_response(self.api))
        self.assertEqual(port['port']['status'], constants.PORT_STATUS_ERROR)
        self.assertEqual(port['port']['name'], 'error_port')

    def test_delete_port_and_network_not_in_nsx(self):
        res = self._create_network('json', 'net1', True)
        net1 = self.deserialize('json', res)
        res = self._create_port('json', net1['network']['id'])
        port = self.deserialize('json', res)
        self.fc._fake_lswitch_dict.clear()
        self.fc._fake_lswitch_lport_dict.clear()
        req = self.new_delete_request('ports', port['port']['id'])
        res = req.get_response(self.api)
        self.assertEqual(res.status_int, 204)
        req = self.new_delete_request('networks', net1['network']['id'])
        res = req.get_response(self.api)
        self.assertEqual(res.status_int, 204)

    def test_delete_router_not_in_nsx(self):
        res = self._create_router('json', 'tenant')
        router = self.deserialize('json', res)
        self.fc._fake_lrouter_dict.clear()
        req = self.new_delete_request('routers', router['router']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 204)

    def test_show_router_not_in_nsx(self):
        res = self._create_router('json', 'tenant')
        router = self.deserialize('json', res)
        self.fc._fake_lrouter_dict.clear()
        req = self.new_show_request('routers', router['router']['id'],
                                    fields=['id', 'status'])
        router = self.deserialize('json', req.get_response(self.ext_api))
        self.assertEqual(router['router']['status'],
                         constants.NET_STATUS_ERROR)

    def _create_network_and_subnet(self, cidr, external=False):
        net_res = self._create_network('json', 'ext_net', True)
        net = self.deserialize('json', net_res)
        net_id = net['network']['id']
        if external:
            self._update('networks', net_id,
                         {'network': {extnet_apidef.EXTERNAL: True}})
        sub_res = self._create_subnet('json', net_id, cidr)
        sub = self.deserialize('json', sub_res)
        return net_id, sub['subnet']['id']

    def test_clear_gateway_nat_rule_not_in_nsx(self):
        # Create external network and subnet
        ext_net_id = self._create_network_and_subnet('1.1.1.0/24', True)[0]
        # Create internal network and subnet
        int_sub_id = self._create_network_and_subnet('10.0.0.0/24')[1]
        res = self._create_router('json', 'tenant')
        router = self.deserialize('json', res)
        # Add interface to router (needed to generate NAT rule)
        req = self.new_action_request(
            'routers',
            {'subnet_id': int_sub_id},
            router['router']['id'],
            "add_router_interface")
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)
        # Set gateway for router
        req = self.new_update_request(
            'routers',
            {'router': {'external_gateway_info':
                        {'network_id': ext_net_id}}},
            router['router']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)
        # Delete NAT rule from NSX, clear gateway
        # and verify operation still succeeds
        self.fc._fake_lrouter_nat_dict.clear()
        req = self.new_update_request(
            'routers',
            {'router': {'external_gateway_info': {}}},
            router['router']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)

    def _test_remove_router_interface_nsx_out_of_sync(self, unsync_action):
        # Create external network and subnet
        ext_net_id = self._create_network_and_subnet('1.1.1.0/24', True)[0]
        # Create internal network and subnet
        int_sub_id = self._create_network_and_subnet('10.0.0.0/24')[1]
        res = self._create_router('json', 'tenant')
        router = self.deserialize('json', res)
        # Set gateway and add interface to router (needed to generate NAT rule)
        req = self.new_update_request(
            'routers',
            {'router': {'external_gateway_info':
                        {'network_id': ext_net_id}}},
            router['router']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)
        req = self.new_action_request(
            'routers',
            {'subnet_id': int_sub_id},
            router['router']['id'],
            "add_router_interface")
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)
        unsync_action()
        req = self.new_action_request(
            'routers',
            {'subnet_id': int_sub_id},
            router['router']['id'],
            "remove_router_interface")
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 200)

    def test_remove_router_interface_not_in_nsx(self):

        def unsync_action():
            self.fc._fake_lrouter_dict.clear()
            self.fc._fake_lrouter_nat_dict.clear()

        self._test_remove_router_interface_nsx_out_of_sync(unsync_action)

    def test_remove_router_interface_nat_rule_not_in_nsx(self):
        self._test_remove_router_interface_nsx_out_of_sync(
            self.fc._fake_lrouter_nat_dict.clear)

    def test_remove_router_interface_duplicate_nat_rules_in_nsx(self):

        def unsync_action():
            # duplicate every entry in the nat rule dict
            tmp = copy.deepcopy(self.fc._fake_lrouter_nat_dict)
            for (_rule_id, rule) in tmp.items():
                self.fc._fake_lrouter_nat_dict[uuid.uuid4()] = rule

        self._test_remove_router_interface_nsx_out_of_sync(unsync_action)

    def test_update_router_not_in_nsx(self):
        res = self._create_router('json', 'tenant')
        router = self.deserialize('json', res)
        self.fc._fake_lrouter_dict.clear()
        req = self.new_update_request(
            'routers',
            {'router': {'name': 'goo'}},
            router['router']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 500)
        req = self.new_show_request('routers', router['router']['id'])
        router = self.deserialize('json', req.get_response(self.ext_api))
        self.assertEqual(router['router']['status'],
                         constants.NET_STATUS_ERROR)

    def test_delete_security_group_not_in_nsx(self):
        res = self._create_security_group('json', 'name', 'desc')
        sec_group = self.deserialize('json', res)
        self.fc._fake_securityprofile_dict.clear()
        req = self.new_delete_request(
            'security-groups',
            sec_group['security_group']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, 204)


class DHCPOptsTestCase(test_dhcpopts.TestExtraDhcpOpt, NsxPluginV2TestCase):

    def setUp(self, plugin=None):
        super(test_dhcpopts.ExtraDhcpOptDBTestCase, self).setUp(
            plugin=vmware.PLUGIN_NAME)



# Copyright 2016 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import xml.etree.ElementTree as et

from networking_sfc.extensions import flowclassifier
from networking_sfc.services.flowclassifier.common import exceptions as exc
from networking_sfc.services.flowclassifier.drivers import base as fc_driver
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as n_context
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from vmware_nsx._i18n import _
from vmware_nsx.common import config  # noqa
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import locking
from vmware_nsx.common import nsxv_constants
from vmware_nsx.plugins.nsx_v.vshield import vcns as nsxv_api
from vmware_nsx.plugins.nsx_v.vshield import vcns_driver
from vmware_nsx.services.flowclassifier.nsx_v import utils as fc_utils

LOG = logging.getLogger(__name__)

REDIRECT_FW_SECTION_NAME = 'OS Flow Classifier Rules'


class NsxvFlowClassifierDriver(fc_driver.FlowClassifierDriverBase):
    """FlowClassifier Driver For NSX-V."""

    _redirect_section_id = None

    def initialize(self):
        self._nsxv = vcns_driver.VcnsDriver(None)
        self.init_profile_id()
        self.init_security_group()
        self.init_security_group_in_profile()

        # register an event to the end of the init to handle the first upgrade
        if self._is_new_security_group:
            registry.subscribe(self.init_complete,
                               resources.PROCESS,
                               events.BEFORE_SPAWN)

    def init_profile_id(self):
        """Init the service insertion profile ID

        Initialize the profile id that should be assigned to the redirect
        rules from the nsx configuration and verify that it exists on backend.
        """
        if not cfg.CONF.nsxv.service_insertion_profile_id:
            raise cfg.RequiredOptError("service_insertion_profile_id",
                                       group=cfg.OptGroup('nsxv'))
        self._profile_id = cfg.CONF.nsxv.service_insertion_profile_id

        # Verify that this moref exists
        if not self._nsxv.vcns.validate_inventory(self._profile_id):
            error = (_("Configured service profile ID: %s not found") %
                     self._profile_id)
            raise nsx_exc.NsxPluginException(err_msg=error)

    def init_security_group(self):
        """Init the service insertion security group

        Look for the service insertion security group in the backend.
        If it was not found - create it
        This security group will contain all the VMs vnics that should
        be inspected by the redirect rules
        """
        # check if this group exist, and create it if not.
        sg_name = fc_utils.SERVICE_INSERTION_SG_NAME
        sg_id = self._nsxv.vcns.get_security_group_id(sg_name)
        self._is_new_security_group = False
        if not sg_id:
            description = ("OpenStack Service Insertion Security Group, "
                           "managed by Neutron nsx-v plugin.")
            sg = {"securitygroup": {"name": sg_name,
                                    "description": description}}
            h, sg_id = (
                self._nsxv.vcns.create_security_group(sg))
            self._is_new_security_group = True

        self._security_group_id = sg_id

    def init_security_group_in_profile(self):
        """Attach the security group to the service profile
        """
        data = self._nsxv.vcns.get_service_insertion_profile(self._profile_id)
        if data and len(data) > 1:
            profile = et.fromstring(data[1])
            profile_binding = profile.find('serviceProfileBinding')
            sec_groups = profile_binding.find('securityGroups')
            for sec in sec_groups.iter('string'):
                if sec.text == self._security_group_id:
                    # Already there
                    return
            # add the security group to the binding
            et.SubElement(sec_groups, 'string').text = self._security_group_id
            self._nsxv.vcns.update_service_insertion_profile_binding(
                self._profile_id,
                et.tostring(profile_binding, encoding="us-ascii"))

    def init_complete(self, resource, event, trigger, payload=None):
        if self._is_new_security_group:
            # add existing VMs to the new security group
            # This code must run after init is done
            core_plugin = directory.get_plugin()
            core_plugin.add_vms_to_service_insertion(
                self._security_group_id)

            # Add the first flow classifier entry
            if cfg.CONF.nsxv.service_insertion_redirect_all:
                self.add_any_any_redirect_rule()

    def add_any_any_redirect_rule(self):
        """Add an any->any flow classifier entry

        Add 1 flow classifier entry that will redirect all the traffic to the
        security partner
        The user will be able to delete/change it later
        """
        context = n_context.get_admin_context()
        fc_plugin = directory.get_plugin(flowclassifier.FLOW_CLASSIFIER_EXT)
        # first check that there is no other flow classifier entry defined:
        fcs = fc_plugin.get_flow_classifiers(context)
        if len(fcs) > 0:
            return

        # Create any->any rule
        fc = {'name': 'redirect_all',
              'description': 'Redirect all traffic',
              'tenant_id': nsxv_constants.INTERNAL_TENANT_ID,
              'l7_parameters': {},
              'ethertype': 'IPv4',
              'protocol': None,
              'source_port_range_min': None,
              'source_port_range_max': None,
              'destination_port_range_min': None,
              'destination_port_range_max': None,
              'source_ip_prefix': None,
              'destination_ip_prefix': None,
              'logical_source_port': None,
              'logical_destination_port': None
              }
        fc_plugin.create_flow_classifier(context, {'flow_classifier': fc})

    def get_redirect_fw_section_id(self):
        if not self._redirect_section_id:
            # try to find it
            self._redirect_section_id = self._nsxv.vcns.get_section_id(
                REDIRECT_FW_SECTION_NAME)
            if not self._redirect_section_id:
                # create it for the first time
                section = et.Element('section')
                section.attrib['name'] = REDIRECT_FW_SECTION_NAME
                self._nsxv.vcns.create_redirect_section(et.tostring(section))
                self._redirect_section_id = self._nsxv.vcns.get_section_id(
                    REDIRECT_FW_SECTION_NAME)

        return self._redirect_section_id

    def get_redirect_fw_section_uri(self):
        return '%s/%s/%s' % (nsxv_api.FIREWALL_PREFIX,
                             nsxv_api.FIREWALL_REDIRECT_SEC_TYPE,
                             self.get_redirect_fw_section_id())

    def get_redirect_fw_section_from_backend(self):
        section_uri = self.get_redirect_fw_section_uri()
        section_resp = self._nsxv.vcns.get_section(section_uri)
        if section_resp and len(section_resp) > 1:
            xml_section = section_resp[1]
            return et.fromstring(xml_section)

    def update_redirect_section_in_backed(self, section):
        section_uri = self.get_redirect_fw_section_uri()
        self._nsxv.vcns.update_section(
            section_uri,
            et.tostring(section, encoding="us-ascii"),
            None)

    def _rule_ip_type(self, flow_classifier):
        if flow_classifier.get('ethertype') == 'IPv6':
            return 'Ipv6Address'
        return 'Ipv4Address'

    def _rule_ports(self, type, flow_classifier):
        min_port = flow_classifier.get(type + '_port_range_min')
        max_port = flow_classifier.get(type + '_port_range_max')
        return self._ports_list(min_port, max_port)

    def _ports_list(self, min_port, max_port):
        """Return a string representing the port/range"""
        if min_port == max_port:
            return str(min_port)
        return "%s-%s" % (min_port, max_port)

    def _rule_name(self, flow_classifier):
        # The name of the rule will include the name & id of the classifier
        # so we can later find it in order to update/delete it.
        # Both the flow classifier DB & the backend has max name length of 255
        # so we may have to trim the name a bit
        return (flow_classifier.get('name')[:200] + '-' +
                flow_classifier.get('id'))

    def _is_the_same_rule(self, rule, flow_classifier_id):
        return rule.find('name').text.endswith(flow_classifier_id)

    def init_redirect_fw_rule(self, redirect_rule, flow_classifier):
        et.SubElement(redirect_rule, 'name').text = self._rule_name(
            flow_classifier)
        et.SubElement(redirect_rule, 'action').text = 'redirect'
        et.SubElement(redirect_rule, 'direction').text = 'inout'
        si_profile = et.SubElement(redirect_rule, 'siProfile')
        et.SubElement(si_profile, 'objectId').text = self._profile_id

        et.SubElement(redirect_rule, 'packetType').text = flow_classifier.get(
            'ethertype').lower()

        # init the source & destination
        if flow_classifier.get('source_ip_prefix'):
            sources = et.SubElement(redirect_rule, 'sources')
            sources.attrib['excluded'] = 'false'
            source = et.SubElement(sources, 'source')
            et.SubElement(source, 'type').text = self._rule_ip_type(
                flow_classifier)
            et.SubElement(source, 'value').text = flow_classifier.get(
                'source_ip_prefix')

        if flow_classifier.get('destination_ip_prefix'):
            destinations = et.SubElement(redirect_rule, 'destinations')
            destinations.attrib['excluded'] = 'false'
            destination = et.SubElement(destinations, 'destination')
            et.SubElement(destination, 'type').text = self._rule_ip_type(
                flow_classifier)
            et.SubElement(destination, 'value').text = flow_classifier.get(
                'destination_ip_prefix')

        # init the service
        if (flow_classifier.get('destination_port_range_min') or
            flow_classifier.get('source_port_range_min')):
            services = et.SubElement(redirect_rule, 'services')
            service = et.SubElement(services, 'service')
            et.SubElement(service, 'isValid').text = 'true'
            if flow_classifier.get('source_port_range_min'):
                source_port = et.SubElement(service, 'sourcePort')
                source_port.text = self._rule_ports('source',
                                                    flow_classifier)

            if flow_classifier.get('destination_port_range_min'):
                dest_port = et.SubElement(service, 'destinationPort')
                dest_port.text = self._rule_ports('destination',
                                                  flow_classifier)

            prot = et.SubElement(service, 'protocolName')
            prot.text = flow_classifier.get('protocol').upper()

        # Add the classifier description
        if flow_classifier.get('description'):
            notes = et.SubElement(redirect_rule, 'notes')
            notes.text = flow_classifier.get('description')

    def _loc_fw_section(self):
        return locking.LockManager.get_lock('redirect-fw-section')

    @log_helpers.log_method_call
    def create_flow_classifier(self, context):
        """Create a redirect rule at the backend
        """
        flow_classifier = context.current
        with self._loc_fw_section():
            section = self.get_redirect_fw_section_from_backend()
            new_rule = et.SubElement(section, 'rule')
            self.init_redirect_fw_rule(new_rule, flow_classifier)
            self.update_redirect_section_in_backed(section)

    @log_helpers.log_method_call
    def update_flow_classifier(self, context):
        """Update the backend redirect rule
        """
        flow_classifier = context.current

        with self._loc_fw_section():
            section = self.get_redirect_fw_section_from_backend()
            redirect_rule = None
            for rule in section.iter('rule'):
                if self._is_the_same_rule(rule, flow_classifier['id']):
                    redirect_rule = rule
                    break

            if redirect_rule is None:
                msg = _("Failed to find redirect rule %s "
                        "on backed") % flow_classifier['id']
                raise exc.FlowClassifierException(message=msg)
            else:
                # The flowclassifier plugin currently supports updating only
                # name or description
                name = redirect_rule.find('name')
                name.text = self._rule_name(flow_classifier)
                notes = redirect_rule.find('notes')
                notes.text = flow_classifier.get('description') or ''
                self.update_redirect_section_in_backed(section)

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context):
        """Delete the backend redirect rule
        """
        flow_classifier_id = context.current['id']
        with self._loc_fw_section():
            section = self.get_redirect_fw_section_from_backend()
            redirect_rule = None
            for rule in section.iter('rule'):
                if self._is_the_same_rule(rule, flow_classifier_id):
                    redirect_rule = rule
                    section.remove(redirect_rule)
                    break

            if redirect_rule is None:
                LOG.error("Failed to delete redirect rule %s: "
                          "Could not find rule on backed",
                          flow_classifier_id)
                # should not fail the deletion
            else:
                self.update_redirect_section_in_backed(section)

    @log_helpers.log_method_call
    def create_flow_classifier_precommit(self, context):
        """Validate the flow classifier data before committing the transaction

        The NSX-v redirect rules does not support:
        - logical ports
        - l7 parameters
        - source ports range / destination port range with more than 15 ports
        """
        flow_classifier = context.current

        # Logical source port
        logical_source_port = flow_classifier['logical_source_port']
        if logical_source_port is not None:
            msg = _('The NSXv driver does not support setting '
                    'logical source port in FlowClassifier')
            raise exc.FlowClassifierBadRequest(message=msg)

        # Logical destination port
        logical_destination_port = flow_classifier['logical_destination_port']
        if logical_destination_port is not None:
            msg = _('The NSXv driver does not support setting '
                    'logical destination port in FlowClassifier')
            raise exc.FlowClassifierBadRequest(message=msg)

        # L7 parameters
        l7_params = flow_classifier['l7_parameters']
        if l7_params is not None and len(l7_params.keys()) > 0:
            msg = _('The NSXv driver does not support setting '
                    'L7 parameters in FlowClassifier')
            raise exc.FlowClassifierBadRequest(message=msg)
